from distutils.version import LooseVersion
import functools
import traceback

import torch
import torchbearer

import sys
if sys.version_info[0] < 3:
    def set_doc(inner, doc):
        return None  # Not simple to do in Python 2.7 so we can leave it for now, just build docs with Python 3+
else:
    def set_doc(inner, doc):
        inner.__doc__ = doc


class no_grad(torch.no_grad):
    """ Context-manager and decorator that disables gradient calculation.
    See `torch.autograd.no_grad <https://pytorch.org/docs/stable/autograd.html#torch.autograd.no_grad>`_
    """
    def __init__(self):
        super(no_grad, self).__init__()
        version = torch.__version__ if str(torch.__version__) is torch.__version__ else "0.4.1"
        if LooseVersion(version) < LooseVersion("0.4.1"):  # No grad is not a decorator
            _patch_call(self, self.call)

    def call(self, func):
        @functools.wraps(func)
        def decorate_no_grad(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_no_grad


def _patch_call(instance, func):
    class _(type(instance)):
        def __call__(self, *arg, **kwarg):
            return func(*arg, **kwarg)
    instance.__class__ = _


class enable_grad(torch.enable_grad):
    """ Context-manager and decorator that enables gradient calculation.
    See `torch.autograd.enable_grad <https://pytorch.org/docs/stable/autograd.html#torch.autograd.enable_grad>`_
    """
    def __init__(self):
        super(enable_grad, self).__init__()
        version = torch.__version__ if str(torch.__version__) is torch.__version__ else "0.4.1"
        if LooseVersion(version) < LooseVersion("0.4.1"):  # Enable grad is not a decorator
            _patch_call(self, self.call)

    def call(self, func):
        @functools.wraps(func)
        def decorate_enable_grad(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_enable_grad


class Metric(object):
    """Base metric class. Process will be called on each batch, process-final at the end of each epoch.
    The metric contract allows for metrics to take any args but not kwargs. The initial metric call will be given state,
    however, subsequent metrics can pass any values desired.

    .. note::

        All metrics must extend this class.

    Args:
        name (str): The name of the metric
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def process(self, *args):
        """Process the state and update the metric for one iteration.

        Args:
            args: Arguments given to the metric. If this is a root level metric, will be given state

        Returns:
            None, or the value of the metric for this batch
        """
        pass

    def process_final(self, *args):
        """Process the terminal state and output the final value of the metric.

        Args:
            args: Arguments given to the metric. If this is a root level metric, will be given state

        Returns:
            None or the value of the metric for this epoch
        """
        pass

    def eval(self, data_key=None):
        """Put the metric in eval mode during model validation.
        """
        pass

    def train(self):
        """Put the metric in train mode during model training.
        """
        pass

    def reset(self, state):
        """Reset the metric, called before the start of an epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass


class Callback(object):
    """Base callback class.

    .. note::

        All callbacks should override this class.

    """

    def state_dict(self):
        """Get a dict containing the callback state.

        Returns:
            dict: A dict containing parameters and persistent buffers.
        """
        return {}

    def __str__(self):
        return str(self.__class__).replace('<class ', '').replace('>', '').replace("'", "")

    def load_state_dict(self, state_dict):
        """Resume this callback from the given state. Expects that this callback was constructed in the same way.

        Args:
            state_dict (dict): The state dict to reload

        Returns:
            Callback: self
        """
        return self

    def on_init(self, state):
        """Perform some action with the given state as context at the init of a trial instance

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start(self, state):
        """Perform some action with the given state as context at the start of a model fit.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_epoch(self, state):
        """Perform some action with the given state as context at the start of each epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_training(self, state):
        """Perform some action with the given state as context at the start of the training loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_sample(self, state):
        """Perform some action with the given state as context after data has been sampled from the generator.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_forward(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_criterion(self, state):
        """Perform some action with the given state as context after the criterion has been evaluated.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_backward(self, state):
        """Perform some action with the given state as context after backward has been called on the loss.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_step_training(self, state):
        """Perform some action with the given state as context after step has been called on the optimiser.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_training(self, state):
        """Perform some action with the given state as context after the training loop has completed.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_start_validation(self, state):
        """Perform some action with the given state as context at the start of the validation loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_sample_validation(self, state):
        """Perform some action with the given state as context after data has been sampled from the validation generator.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_forward_validation(self, state):
        """Perform some action with the given state as context after the forward pass (model output) has been completed
        with the validation data.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_criterion_validation(self, state):
        """Perform some action with the given state as context after the criterion evaluation has been completed
        with the validation data.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_step_validation(self, state):
        """Perform some action with the given state as context at the end of each validation step.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_validation(self, state):
        """Perform some action with the given state as context at the end of the validation loop.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end_epoch(self, state):
        """Perform some action with the given state as context at the end of each epoch.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_checkpoint(self, state):
        """Perform some action with the state after all other callbacks have completed at the end of an epoch and the
        history has been updated. Should only be used for taking checkpoints or snapshots and will only be called by the
        run method of Trial.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass

    def on_end(self, state):
        """Perform some action with the given state as context at the end of the model fitting.

        Args:
            state (dict): The current state dict of the :class:`.Trial`.
        """
        pass


def _get_param_list(param):
    if isinstance(param, list):
        return param
    if isinstance(param, tuple):
        return list(param)
    return [param]


def _forward_with_exceptions(x, model, y_pred, state):
    dx = state[x]

    # Forward Pass
    try:
        exc_info = sys.exc_info()
        state[y_pred] = state[model](*_get_param_list(dx), state=state)
    except Exception as e:
        error = []
        try:
            state[y_pred] = state[model](*_get_param_list(dx))
        except TypeError as e2:
            if isinstance(e, TypeError):  # If both are type errors, show both.
                error.append(e2)
            error.append(e)
            raise Exception(error)
        except Exception as e2:
            if not isinstance(e, TypeError):
                error.append(e)
            error.append(e2)
            raise Exception(error)
    finally:
        print_trace = False
        for exc in exc_info:
            if exc is not None:
                print_trace = True

        traceback.print_exception(*exc_info) if print_trace else None


def base_closure(x, model, y_pred, y_true, crit, loss, opt):
    """Function to create a standard pytorch closure using objects taken from state under the given keys.

    Args:
        x: State key under which the input data is stored
        model: State key under which the pytorch model is stored
        y_pred: State key under which the predictions will be stored
        y_true: State key under which the targets are stored
        crit: State key under which the criterion function is stored (function of state or (y_pred, y_true))
        loss: State key under which the loss will be stored
        opt: State key under which the optimsiser is stored

    Returns:
        function: Standard closure function
    """
    def closure(state):
        # Zero grads
        state[opt].zero_grad()

        _forward_with_exceptions(x, model, y_pred, state)

        state[torchbearer.CALLBACK_LIST].on_forward(state)

        # Loss Calculation
        try:
            state[loss] = state[crit](state)
        except TypeError:
            loss_function_params = _get_param_list(state[y_pred]) + _get_param_list(state[y_true])
            state[loss] = state[crit](*loss_function_params)

        state[torchbearer.CALLBACK_LIST].on_criterion(state)

        # Backwards pass
        state[loss].backward(**state[torchbearer.BACKWARD_ARGS])

        state[torchbearer.CALLBACK_LIST].on_backward(state)
    return closure


standard_closure = lambda: base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                                torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)


def apex_closure():
    from apex import amp

    def _apex_closure(state):
        # Zero grads
        state[torchbearer.OPTIMIZER].zero_grad()

        _forward_with_exceptions(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, state)

        state[torchbearer.CALLBACK_LIST].on_forward(state)

        # Loss Calculation
        try:
            state[torchbearer.LOSS] = state[torchbearer.CRITERION](state)
        except TypeError:
            loss_function_params = _get_param_list(state[torchbearer.Y_PRED]) + _get_param_list(state[torchbearer.Y_TRUE])
            state[torchbearer.LOSS] = state[torchbearer.CRITERION](*loss_function_params)

        state[torchbearer.CALLBACK_LIST].on_criterion(state)

        # Backwards pass
        with amp.scale_loss(state[torchbearer.LOSS], state[torchbearer.OPTIMIZER]) as scaled_loss:
            scaled_loss.backward(**state[torchbearer.BACKWARD_ARGS])

        state[torchbearer.CALLBACK_LIST].on_backward(state)
    return _apex_closure


def cite(bibtex):
    """A decorator which adds a reference to the **Google style** docstring of the given object. The ``Args:`` or
    ``Returns:`` line is then prepended with the given bibtex string at runtime. Otherwise, the last line is used.

    Args:
        bibtex (str): The bibtex string to insert

    Returns:
        The decorator
    """
    def decorator(inner):
        doc = inner.__doc__.split('\n')
        i = 0
        s = 0
        for line in doc:
            sline = line.strip()
            if sline == 'Args:' or sline == 'Returns:':
                for char in line:
                    if char == ' ':
                        s += 1
                break
            i += 1

        spaces = ' ' * (s + 4)
        to_insert = ' ' * s + '::\n\n' + spaces
        to_insert += bibtex.strip().replace('\n', '\n' + spaces).rstrip()

        doc.insert(i, '')
        doc.insert(i, to_insert)
        set_doc(inner, '\n'.join(doc))
        return inner
    return decorator



from unittest import TestCase

import torch

import torchbearer
from torchbearer.callbacks.cutout import Cutout, RandomErase, CutMix


class TestCutOut(TestCase):
    def test_cutout(self):
        random_image = torch.rand(2, 3, 100, 100)
        torch.manual_seed(7)
        co = Cutout(1, 10)
        state = {torchbearer.X: random_image}
        co.on_sample(state)
        reg_img = state[torchbearer.X].view(-1)

        x = [21, 86]
        y = [15, 92]

        known_cut = random_image.clone().numpy()
        known_cut[0, :, y[0]-10//2:y[0]+10//2, x[0]-10//2:x[0]+10//2] = 0
        known_cut[1, :, y[1]-10//2:y[1]+10//2, x[1]-10//2:x[1]+10//2] = 0
        known_cut = torch.from_numpy(known_cut)
        known_cut = known_cut.view(-1)

        diff = (torch.abs(known_cut-reg_img) > 1e-4).any()
        self.assertTrue(diff.item() == 0)

    def test_cutout_constant(self):
        random_image = torch.rand(2, 3, 100, 100)
        torch.manual_seed(7)
        co = Cutout(1, 10, constant=0.5)
        state = {torchbearer.X: random_image}
        co.on_sample(state)
        reg_img = state[torchbearer.X].view(-1)

        x = [21, 86]
        y = [15, 92]

        known_cut = random_image.clone().numpy()
        known_cut[0, :, y[0]-10//2:y[0]+10//2, x[0]-10//2:x[0]+10//2] = 0.5
        known_cut[1, :, y[1]-10//2:y[1]+10//2, x[1]-10//2:x[1]+10//2] = 0.5
        known_cut = torch.from_numpy(known_cut)
        known_cut = known_cut.view(-1)

        diff = (torch.abs(known_cut-reg_img) > 1e-4).any()
        self.assertTrue(diff.item() == 0)

    # TODO: Find a better test for this
    def test_random_erase(self):
        random_image = torch.rand(2, 3, 100, 100)
        torch.manual_seed(7)
        co = RandomErase(1, 10)
        state = {torchbearer.X: random_image}
        co.on_sample(state)
        reg_img = state[torchbearer.X].view(-1)

        x = [21, 86]
        y = [15, 92]

        known_cut = random_image.clone().numpy()
        known_cut[0, :, y[0]-10//2:y[0]+10//2, x[0]-10//2:x[0]+10//2] = 0
        known_cut[1, :, y[1]-10//2:y[1]+10//2, x[1]-10//2:x[1]+10//2] = 0
        known_cut = torch.from_numpy(known_cut)

        known_cut = known_cut.view(-1)
        masked_pix = known_cut == 0

        diff = (torch.abs(known_cut[masked_pix]-reg_img[masked_pix]) > 1e-4).any()
        self.assertTrue(diff.item() > 0)

    def test_cutmix(self):
        random_image = torch.rand(5, 3, 100, 100)
        state = {torchbearer.X: random_image, torchbearer.Y_TRUE: torch.randint(10, (5,)).long(), torchbearer.DEVICE: 'cpu'}
        torch.manual_seed(7)
        co = CutMix(0.25, classes=10)
        co.on_sample(state)
        reg_img = state[torchbearer.X].view(-1)

        x = [72, 83, 18, 96, 40]
        y = [8, 17, 62, 30, 66]
        perm = [0, 4, 3, 2, 1]
        sz = 3

        rnd = random_image.clone().numpy()
        known_cut = random_image.clone().numpy()
        known_cut[0, :, y[0]-sz//2:y[0]+sz//2, x[0]-sz//2:x[0]+sz//2] = rnd[perm[0], :, y[0]-sz//2:y[0]+sz//2, x[0]-sz//2:x[0]+sz//2]
        known_cut[1, :, y[1]-sz//2:y[1]+sz//2, x[1]-sz//2:x[1]+sz//2] = rnd[perm[1], :, y[1]-sz//2:y[1]+sz//2, x[1]-sz//2:x[1]+sz//2]
        known_cut[2, :, y[2]-sz//2:y[2]+sz//2, x[2]-sz//2:x[2]+sz//2] = rnd[perm[2], :, y[2]-sz//2:y[2]+sz//2, x[2]-sz//2:x[2]+sz//2]
        known_cut[3, :, y[3]-sz//2:y[3]+sz//2, x[3]-sz//2:x[3]+sz//2] = rnd[perm[3], :, y[3]-sz//2:y[3]+sz//2, x[3]-sz//2:x[3]+sz//2]
        known_cut[4, :, y[4]-sz//2:y[4]+sz//2, x[4]-sz//2:x[4]+sz//2] = rnd[perm[4], :, y[4]-sz//2:y[4]+sz//2, x[4]-sz//2:x[4]+sz//2]
        known_cut = torch.from_numpy(known_cut)
        known_cut = known_cut.view(-1)

        diff = (torch.abs(known_cut-reg_img) > 1e-4).any()
        self.assertTrue(diff.item() == 0)

    def test_cutmix_targets(self):
        random_image = torch.rand(2, 3, 100, 100)
        torch.manual_seed(7)
        co = CutMix(1.0, classes=4)
        target = torch.tensor([
            [0., 1., 0., 0.],
            [0., 0., 0., 1.]
        ])
        state = {torchbearer.X: random_image, torchbearer.Y_TRUE: torch.tensor([1, 3]).long(), torchbearer.DEVICE: 'cpu'}
        co.on_sample(state)
        self.assertTrue(((state[torchbearer.TARGET] - target).abs() < 0.00001).all())
        state = {torchbearer.X: random_image, torchbearer.Y_TRUE: torch.tensor([1, 3]).long()}
        co.on_sample_validation(state)
        self.assertTrue(((state[torchbearer.TARGET] - target).abs() < 0.00001).all())
        state = {torchbearer.X: random_image, torchbearer.Y_TRUE: target.long()}
        co.on_sample_validation(state)
        self.assertTrue(((state[torchbearer.TARGET] - target).abs() < 0.00001).all())



"""
Aggregators are a special kind of :class:`.Metric` which takes as input, the output from a previous metric or metrics.
As a result, via a :class:`.MetricTree`, a series of aggregators can collect statistics such as Mean or Standard
Deviation without needing to compute the underlying metric multiple times. This can, however, make the aggregators
complex to use. It is therefore typically better to use the :mod:`decorator API<.metrics.decorators>`.
"""
from collections import deque

import torch

from torchbearer.bases import Metric
from .metrics import AdvancedMetric


class RunningMetric(AdvancedMetric):
    """A metric which aggregates batches of results and presents a method to periodically process these into a value.

    .. note::

       Running metrics only provide output during training.

    Args:
        name (str): The name of the metric.
        batch_size (int): The size of the deque to store of previous results.
        step_size (int): The number of iterations between aggregations.
    """
    def __init__(self, name, batch_size=50, step_size=10):
        super(RunningMetric, self).__init__(name)
        self._batch_size = batch_size
        self._step_size = step_size
        self._cache = deque()
        self._result = {}

    def _process_train(self, *args):
        """Process the metric for a single train step.

        Args:
            args: The output of some :class:`.Metric`

        Returns:
            The metric value.
        """
        raise NotImplementedError

    def _step(self, cache):
        """Aggregate the cache to produce a single metric value.

        Args:
            cache (list): The current stored metric cache.

        Returns:
            The new metric value.
        """
        raise NotImplementedError

    def process_train(self, *args):
        """Add the current metric value to the cache and call '_step' is needed.

        Args:
            args: The output of some :class:`.Metric`

        Returns:
            The current metric value.
        """
        self._cache.append(self._process_train(*args))
        if len(self._cache) > self._batch_size:
            self._cache.popleft()
        if self._i % self._step_size == 0:
            self._result = self._step(list(self._cache))
        self._i += 1
        return self._result

    def reset(self, state):
        """Reset the step counter. Does not clear the cache.

        Args:
            state (dict): The current model state.
        """
        self._i = 0


class RunningMean(RunningMetric):
    """A :class:`RunningMetric` which outputs the running mean of its input tensors over the course of an epoch.

    Args:
        name (str): The name of this running mean.
        batch_size (int): The size of the deque to store of previous results.
        step_size (int): The number of iterations between aggregations.
        dim (int, tuple): The dimension(s) on which to perform the mean. If left as None, this will mean over the whole
            Tensor
    """

    def __init__(self, name, batch_size=50, step_size=10, dim=None):
        super(RunningMean, self).__init__(name, batch_size=batch_size, step_size=step_size)
        self._kwargs = {'dim': dim} if dim is not None else {}

    def _process_train(self, *args):
        data = args[0]
        res = data.mean(**self._kwargs)
        return res

    def _step(self, cache):
        res = sum(cache) / float(len(cache))
        return res.item() if res.numel() <= 1 else res.tolist()


class Var(Metric):
    """Metric aggregator which calculates the **sample** variance of process outputs between calls to reset.
    Optionally calculate the population variance if ``unbiased = False``.

    Args:
        name (str): The name of this metric.
        unbiased (bool): If True (default), calculates the sample variance, else, the population variance
        dim (int, tuple): The dimension(s) on which to compute the std. If left as None, this will operate over the
            whole Tensor
    """

    def __init__(self, name, unbiased=True, dim=None):
        super(Var, self).__init__(name)
        self._unbiased = unbiased
        self._kwargs = {'dim': dim} if dim is not None else {}
        self._sum = 0.0
        self._sum_sq = 0.0
        self._count = 0

    def process(self, *args):
        """Compute values required for the variance from the input. The input should be a torch Tensor. The sum and sum
        of squares will be computed over the provided dimension.

        Args:
            args (torch.Tensor):  The output of some previous call to :meth:`.Metric.process`.
        """
        data = args[0]
        tot = data.sum(**self._kwargs)
        self._sum += tot
        self._sum_sq += data.pow(2).sum(**self._kwargs)
        self._count += data.numel() / tot.numel()

    def _process_final(self):
        mean = self._sum / self._count
        mean = mean.pow(2)
        variance = ((self._sum_sq / self._count) - mean).clamp(min=0)
        if self._unbiased:
            variance = variance * (self._count / (self._count - 1.0))
        return variance

    def process_final(self, *args):
        """Compute and return the final variance.

        Returns:
            The variance of each observation since the last reset call.
        """
        res = self._process_final()
        return res.item() if res.numel() <= 1 else res.tolist()

    def reset(self, state):
        """Reset the statistics to compute the next variance.

        Args:
            state (dict): The model state.
        """
        super(Var, self).reset(state)
        self._sum = 0.0
        self._sum_sq = 0.0
        self._count = 0


class Std(Var):
    """Metric aggregator which calculates the **sample** standard deviation of process outputs between calls to reset.
    Optionally calculate the population std if ``unbiased = False``.

    Args:
        name (str): The name of this metric.
        unbiased (bool): If True (default), calculates the sample standard deviation, else, the population standard
            deviation
        dim (int, tuple): The dimension(s) on which to compute the std. If left as None, this will operate over the
            whole Tensor
    """

    def __init__(self, name, unbiased=True, dim=None):
        super(Std, self).__init__(name, unbiased=unbiased, dim=dim)

    def process_final(self, *args):
        """Compute and return the final standard deviation.

        Returns:
            The standard deviation of each observation since the last reset call.
        """
        res = super(Std, self)._process_final().sqrt()
        return res.item() if res.numel() <= 1 else res.tolist()


class Mean(Metric):
    """Metric aggregator which calculates the mean of process outputs between calls to reset.

    Args:
        name (str): The name of this metric.
        dim (int, tuple): The dimension(s) on which to perform the mean. If left as None, this will mean over the whole
            Tensor
    """

    def __init__(self, name, dim=None):
        super(Mean, self).__init__(name)
        self._kwargs = {'dim': dim} if dim is not None else {}
        self._sum = 0.0
        self._count = 0

    def process(self, *args):
        """Add the input to the rolling sum. Input must be a torch tensor.

        Args:
            args:  The output of some previous call to :meth:`.Metric.process`.
        """
        data = args[0]
        tot = data.sum(**self._kwargs)
        self._sum += tot
        self._count += data.numel() / tot.numel()

    def process_final(self, *args):
        """Compute and return the mean of all metric values since the last call to reset.

        Returns:
            The mean of the metric values since the last call to reset.
        """
        res = self._sum / self._count
        return res.item() if res.numel() <= 1 else res.tolist()

    def reset(self, state):
        """Reset the running count and total.

        Args:
            state (dict): The model state.
        """
        super(Mean, self).reset(state)
        self._sum = 0.0
        self._count = 0



from unittest import TestCase
from mock import patch, Mock

from torch import nn
from torchbearer.callbacks.manifold_mixup import ManifoldMixup
import torchbearer
import torch


class TestModule(nn.Module):
    def __init__(self):
        super(TestModule, self).__init__()
        self.conv = nn.Conv1d(1, 1, 1)
        self.relu = nn.ReLU()
        self.bn = nn.BatchNorm1d(1)

    def forward(self, x):
        x = self.conv(x.view(-1, 1, 1))
        x = self.relu(x)
        x = self.bn(x)
        return x


class TestModule2(nn.Module):
    def __init__(self):
        super(TestModule2, self).__init__()
        self.layer1 = TestModule()

    def forward(self, x):
        return self.layer1(x)


class TestModel(nn.Module):
    def __init__(self):
        super(TestModel, self).__init__()
        self.fc1 = nn.Linear(1, 1)
        self.conv1 = nn.Conv1d(1, 1, 1)
        self.relu = nn.ReLU()
        self.layer1 = TestModule()
        self.layer2 = TestModule2()

    def forward(self, x):
        x = self.fc1(x)
        x = self.conv1(x.view(-1,1,1))
        x = self.relu(x)
        x = self.layer1(x)
        x = self.layer2(x)
        return x


class TestManifoldMixup(TestCase):
    def setUp(self):
        super(TestManifoldMixup, self).setUp()
        self.model = TestModel()

    def test_depth_none(self):
        mm = ManifoldMixup().at_depth(None)
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        self.assertTrue(len(mm._layers) == 12)
        
    def test_depth_0(self):
        mm = ManifoldMixup().at_depth(0)
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        checks = [
            self.model.fc1 in mm._layers,
            self.model.conv1 in mm._layers,
            self.model.relu in mm._layers,
            self.model.layer1 in mm._layers,
            self.model.layer2 in mm._layers,
            ]

        self.assertTrue(all(checks))  # Top level modules in
        self.assertFalse(self.model.layer1.conv in mm._layers)  # Depth 1 modules not in

    def test_depth_1(self):
        mm = ManifoldMixup().at_depth(1)
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        top_checks = [
            self.model.fc1 in mm._layers,
            self.model.conv1 in mm._layers,
            self.model.relu in mm._layers,
            self.model.layer1 in mm._layers,
            self.model.layer2 in mm._layers,
        ]

        first_checks = [
            self.model.layer1.conv in mm._layers,
            self.model.layer1.relu in mm._layers,
            self.model.layer1.bn in mm._layers,
            self.model.layer2.layer1 in mm._layers,
        ]

        self.assertFalse(any(top_checks))  # Top level modules not in
        self.assertTrue(all(first_checks))  # Depth 1 modules in

    def test_depth_2(self):
        mm = ManifoldMixup().at_depth(2)
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        top_checks = [
            self.model.fc1 in mm._layers,
            self.model.conv1 in mm._layers,
            self.model.relu in mm._layers,
            self.model.layer1 in mm._layers,
            self.model.layer2 in mm._layers,
        ]

        first_checks = [
            self.model.layer1.conv in mm._layers,
            self.model.layer1.relu in mm._layers,
            self.model.layer1.bn in mm._layers,
            self.model.layer2.layer1 in mm._layers,
        ]

        second_checks = [
            self.model.layer2.layer1.conv in mm._layers,
            self.model.layer2.layer1.relu in mm._layers,
            self.model.layer2.layer1.bn in mm._layers,
        ]

        self.assertFalse(any(top_checks))  # Top level modules not in
        self.assertFalse(any(first_checks))  # Depth 1 modules not in
        self.assertTrue(all(second_checks))  # Depth 2 modules in

    def test_for_layers(self):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_conv', 'layer2_layer1_conv'])
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)
        
        self.assertTrue(self.model.conv1 in mm._layers, self.model.layer1.conv in mm._layers and self.model.layer2.layer1.conv in mm._layers)
        self.assertTrue(len(mm._layers) == 3)

    def test_get_selected_layers(self):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_conv', 'layer2_layer1_conv'])
        found_layers = mm.get_selected_layers(self.model)
        self.assertTrue(len(found_layers) == 3)
        self.assertTrue('conv1' in found_layers)
        self.assertTrue('layer1_conv' in found_layers)
        self.assertTrue('layer2_layer1_conv' in found_layers)

    def test_layer_filter(self):
        mm = ManifoldMixup().at_depth(None).with_layer_filter(['conv1', 'layer1_conv', 'layer2_layer1_conv'])
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        self.assertFalse(self.model.conv1 in mm._layers,
                        self.model.layer1.conv in mm._layers and self.model.layer2.layer1.conv in mm._layers)
        self.assertTrue(len(mm._layers) == 12-3)

    def test_layer_type_filter(self):
        mm = ManifoldMixup().at_depth(None).with_layer_type_filter([nn.Conv1d])
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        self.assertFalse(self.model.conv1 in mm._layers,
                        self.model.layer1.conv in mm._layers and self.model.layer2.layer1.conv in mm._layers)
        self.assertTrue(len(mm._layers) == 12-3)

    def test_wrap(self):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_relu', 'layer2_layer1_conv'])
        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        self.model.conv1.mixup()
        self.model.layer1.relu.mixup()
        self.model.layer2.layer1.conv.mixup()

        self.assertRaises(AttributeError, lambda: self.model.relu.mixup())

    @patch('torchbearer.callbacks.manifold_mixup._mixup_inputs', side_effect=lambda x, _: x)
    def test_call_mix(self, _):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_relu', 'layer2_layer1_conv'])

        state = {torchbearer.MODEL: self.model}
        mm.on_start(state)

        self.model.conv1.mixup()
        self.assertTrue(self.model.conv1.do_mixup)
        self.model(torch.rand(3, 1))
        self.assertFalse(self.model.conv1.do_mixup)

    @patch('torchbearer.callbacks.manifold_mixup._mixup')
    def test_on_sample(self, mix):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_relu', 'layer2_layer1_conv'])

        state = {torchbearer.MODEL: self.model, torchbearer.X: torch.rand(3, 1), torchbearer.Y_TRUE: torch.rand(3, 1)}
        mm.on_start(state)

        mm.on_sample(state)
        self.assertTrue(mix.call_count == 1)
        
        self.assertTrue(torchbearer.MIXUP_PERMUTATION in state)
        self.assertTrue(torchbearer.MIXUP_LAMBDA in state)

        state = {torchbearer.MODEL: self.model, torchbearer.X: torch.rand(3, 1), torchbearer.Y_TRUE: torch.rand(3, 1)}
        mm.on_sample(state)
        self.assertTrue(mix.call_count == 2)

    @patch('torchbearer.callbacks.manifold_mixup._mixup_inputs', side_effect=lambda x, _: x)
    def test_eval(self, mix):
        mm = ManifoldMixup().at_depth(None).for_layers(['conv1', 'layer1_relu', 'layer2_layer1_conv'])

        self.model.eval()
        state = {torchbearer.MODEL: self.model, torchbearer.X: torch.rand(3, 1), torchbearer.Y_TRUE: torch.rand(3, 1)}
        mm.on_start(state)

        mm.on_sample(state)
        self.model(torch.rand(3, 1))
        self.assertTrue(mix.call_count == 0)

        state = {torchbearer.MODEL: self.model, torchbearer.X: torch.rand(3, 1), torchbearer.Y_TRUE: torch.rand(3, 1)}
        mm.on_sample(state)
        self.model = self.model.train()
        self.model(torch.rand(3, 1))
        self.assertTrue(mix.call_count == 1)

    def test_mixup_inputs(self):
        from torchbearer.callbacks.manifold_mixup import _mixup_inputs

        x = torch.Tensor([[1, 2], [2, 3]])
        perm = torch.Tensor([1, 0]).long()
        lam = torch.Tensor([0.1])

        state = {torchbearer.X: x, torchbearer.MIXUP_PERMUTATION: perm, torchbearer.MIXUP_LAMBDA: lam}
        mixed = _mixup_inputs(x, state)

        self.assertFalse((mixed - torch.Tensor([[1.9, 2.9], [1.1, 2.1]]) > 1e-6).any())

    @patch('torchbearer.callbacks.manifold_mixup.Beta')
    def test_sample_lam_random(self, beta):
        mm = ManifoldMixup()
        sl = mm._sample_lam
        sl()

        self.assertTrue(beta.mock_calls[0][1] == (1., 1.))
        self.assertTrue(beta.mock_calls[1][0] == '().sample')

    def test_sample_lam_negative(self):
        mm = ManifoldMixup(alpha=-1)
        sl = mm._sample_lam
        lam = sl()

        self.assertTrue(lam == 1.)

    def test_sample_lam_fixed(self):
        mm = ManifoldMixup(lam=2.)
        sl = mm._sample_lam
        lam = sl()

        self.assertTrue(lam == 2.)
        
    def test_single_to_list(self):
        mm = ManifoldMixup()
        sl = mm._single_to_list

        item = 1.
        self.assertTrue(sl(item) == [item, ])







# Copyright (C) 2017, Dmytro Mishkin
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This file has been modified from https://github.com/ducha-aiki/LSUV-pytorch

import numpy as np
import torch
import torch.nn.init
import torch.nn as nn


class LSUV(object):
    """Initialisation from the paper `All you need is a good init <https://arxiv.org/abs/1511.06422>`__. Call
    LSUV.init_model(...) on a torch model to perform the initialisation. Implementation based off of the PyTorch
    implementation `here <https://github.com/ducha-aiki/LSUV-pytorch>`__.

    Example: ::

        >>> import torch.nn
        >>> from torchbearer.callbacks import LSUV

        # Example of initialising a torch model with LSUV
        >>> model = torch.nn.Linear(1,1)
        >>> data = torch.rand(100, 1)
        >>> lsuv = LSUV(model, data)
        >>> model = lsuv.init_model(model, data)

    """
    def __init__(self):
        super(LSUV, self).__init__()
        self.gg = self.reset_parameters()

    def reset_parameters(self):
        self.gg = {
            'hook_position': 0,
            'total_fc_conv_layers': 0,
            'done_counter': -1,
            'hook': None,
            'act_dict': {},
            'counter_to_apply_correction': 0,
            'correction_needed': False,
            'current_coef': 1.0,
            'weight_lambda': lambda m: m.weight,
        }
        return self.gg
        
    def svd_orthonormal(self, w):
        shape = w.shape
        flat_shape = (shape[0], np.prod(shape[1:]))
        a = torch.rand(flat_shape, device=w.device)
        u, _, v = torch.svd(a, some=True)
        q = u if u.shape == flat_shape else v.t()
        q = q.view(shape)
        return q.to(torch.float)
    
    def add_current_hook(self, m):
        if self.gg['hook'] is not None:
            return
        if (isinstance(m, nn.Conv2d)) or (isinstance(m, nn.Linear)):
            if self.gg['hook_position'] > self.gg['done_counter']:
                self.gg['hook'] = m.register_forward_hook(self.store_activations_wrapper())
            else:
                self.gg['hook_position'] += 1
    
    def count_conv_fc_layers(self, m):
        if (isinstance(m, nn.Conv2d)) or (isinstance(m, nn.Linear)):
            self.gg['total_fc_conv_layers'] += 1
    
    def orthogonal_weights_init(self, m):
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
            weight = self.gg['weight_lambda'](m)
            w_ortho = self.svd_orthonormal(weight.data)
            m.weight.data = w_ortho.data
    
    def apply_weights_correction(self, m):
        if self.gg['hook'] is None or not self.gg['correction_needed']:
            return
        if (isinstance(m, nn.Conv2d)) or (isinstance(m, nn.Linear)):
            if self.gg['counter_to_apply_correction'] < self.gg['hook_position']:
                self.gg['counter_to_apply_correction'] += 1
            else:
                weight = self.gg['weight_lambda'](m)
                weight.data *= float(self.gg['current_coef'])
                self.gg['correction_needed'] = False
    
    def init_model(self, model, data, weight_lambda=None, needed_std=1.0, std_tol=0.1, max_attempts=10, do_orthonorm=True):
        self.gg = self.reset_parameters()
        train = True if model.training else False
        self.gg['weight_lambda'] = self.gg['weight_lambda'] if weight_lambda is None else weight_lambda
    
        model.eval()
        model.apply(self.count_conv_fc_layers)
        if do_orthonorm:
            model.apply(self.orthogonal_weights_init)
        for layer_idx in range(self.gg['total_fc_conv_layers']):
            model.apply(self.add_current_hook)
            _ = model(data)
            current_std = self.gg['act_dict'].std()
            attempts = 0
            while torch.abs(current_std - needed_std).item() > std_tol:
                self.gg['current_coef'] = needed_std / (current_std + 1e-8)
                self.gg['correction_needed'] = True
                model.apply(self.apply_weights_correction)
                _ = model(data)
                current_std = self.gg['act_dict'].std()
                attempts += 1
                if attempts > max_attempts:
                    break
            if self.gg['hook'] is not None:
                self.gg['hook'].remove()
            self.gg['done_counter'] += 1
            self.gg['counter_to_apply_correction'] = 0
            self.gg['hook_position'] = 0
            self.gg['hook'] = None
              
        if train:
            model.train()
        return model

    def store_activations_wrapper(self):
        gg = self.gg
        def store_activations(self, input, output):
            gg['act_dict'] = output.data
        return store_activations



import sys
if sys.version_info[0] < 3:
    import inspect
    def count_args(fcn):
        return len(inspect.getargspec(fcn).args)
else:
    from inspect import signature
    def count_args(fcn):
        return len(signature(fcn).parameters)

import types

import torchbearer
from torchbearer.callbacks import Callback


class LambdaCallback(Callback):
    def __init__(self, func):
        self.func = func

    def on_lambda(self, state):
        return self.func(state)


def bind_to(target):
    def decorator(func):
        if isinstance(func, LambdaCallback):
            callback = func
        else:
            callback = LambdaCallback(func)
        setattr(callback, target.__name__, types.MethodType(lambda self, state: self.on_lambda(state), callback))
        return callback
    return decorator


def on_init(func):
    """ The :func:`on_init` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_init`
    calling the decorated function

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_init

        # Example callback on start
        >>> @on_init
        ... def print_callback(state):
        ...     print('Initialised trial.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Initialised trial.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_init` calling func
    """
    return bind_to(Callback.on_init)(func)


def on_start(func):
    """ The :func:`on_start` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start`
    calling the decorated function

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_start

        # Example callback on start
        >>> @on_start
        ... def print_callback(state):
        ...     print('Starting training.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Starting training.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start` calling func
    """
    return bind_to(Callback.on_start)(func)


def on_start_epoch(func):
    """ The :func:`on_start_epoch` decorator is used to initialise a :class:`.Callback` with
    :meth:`~.Callback.on_start_epoch` calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_start_epoch

        # Example callback running at start of each epoch
        >>> @on_start_epoch
        ... def print_callback(state):
        ...     print('Starting epoch {}.'.format(state[torchbearer.EPOCH]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Starting epoch 0.

        Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_epoch` calling func
    """
    return bind_to(Callback.on_start_epoch)(func)


def on_start_training(func):
    """ The :func:`on_start_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_training`
    calling the decorated function

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_start_training

        # Example callback running at start of the training pass
        >>> @on_start_training
        ... def print_callback(state):
        ...     print('Starting training.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Starting training.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_training` calling func
    """
    return bind_to(Callback.on_start_training)(func)


def on_sample(func):
    """ The :func:`on_sample` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_sample

        # Example callback running each time a sample is taken from the dataset
        >>> @on_sample
        ... def print_callback(state):
        ...     print('Current sample {}.'.format(state[torchbearer.X]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Current sample None.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_sample` calling func
    """
    return bind_to(Callback.on_sample)(func)


def on_forward(func):
    """ The :func:`on_forward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward`
    calling the decorated function

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_forward

        # Example callback running after each training forward pass of the torch model
        >>> @on_forward
        ... def print_callback(state):
        ...     print('Evaluated training batch.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Evaluated training batch.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_forward` calling func
    """
    return bind_to(Callback.on_forward)(func)


def on_criterion(func):
    """ The :func:`on_criterion` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_criterion

        # Example callback running after each evaluation of the loss
        >>> @on_criterion
        ... def print_callback(state):
        ...     print('Current loss {}.'.format(state[torchbearer.LOSS].item()))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Current loss 0.0.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_criterion` calling func
    """
    return bind_to(Callback.on_criterion)(func)


def on_backward(func):
    """ The :func:`on_backward` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_backward`
    calling the decorated function

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_backward

        # Example callback running after each backward pass of the torch model
        >>> @on_backward
        ... def print_callback(state):
        ...     print('Doing backward.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Doing backward.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_backward` calling func
    """
    return bind_to(Callback.on_backward)(func)


def on_step_training(func):
    """ The :func:`on_step_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_training`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_step_training

        # Example callback running after each training step
        >>> @on_step_training
        ... def print_callback(state):
        ...     print('Step {}.'.format(state[torchbearer.BATCH]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Step 0.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_step_training` calling func
    """
    return bind_to(Callback.on_step_training)(func)


def on_end_training(func):
    """ The :func:`on_end_training` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_training`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_end_training

        # Example callback running after each training pass
        >>> @on_end_training
        ... def print_callback(state):
        ...     print('Finished training pass.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Finished training pass.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_training` calling func
    """
    return bind_to(Callback.on_end_training)(func)


def on_end_epoch(func):
    """ The :func:`on_end_epoch` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_epoch`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_end_epoch

        # Example callback running each epoch
        >>> @on_end_epoch
        ... def print_callback(state):
        ...     print('Finished epoch {}.'.format(state[torchbearer.EPOCH]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Finished epoch 0.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_epoch` calling func
    """
    return bind_to(Callback.on_end_epoch)(func)


def on_end(func):
    """ The :func:`on_end` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_end

        # Example callback running after all training is finished.
        >>> @on_end
        ... def print_callback(state):
        ...     print('Finished training model.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Finished training model.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end` calling func
    """
    return bind_to(Callback.on_end)(func)


def on_start_validation(func):
    """ The :func:`on_start_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_start_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_start_validation

        # Example callback running when each validation pass starts.
        >>> @on_start_validation
        ... def print_callback(state):
        ...     print('Starting validation.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Starting validation.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_start_validation` calling func
    """
    return bind_to(Callback.on_start_validation)(func)


def on_sample_validation(func):
    """ The :func:`on_sample_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_sample_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_sample_validation

        # Example callback running after each validation sample is drawn.
        >>> @on_sample_validation
        ... def print_callback(state):
        ...     print('Sampled validation data {}.'.format(state[torchbearer.X]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Sampled validation data None.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_sample_validation` calling func
    """
    return bind_to(Callback.on_sample_validation)(func)


def on_forward_validation(func):
    """ The :func:`on_forward_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_forward_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_forward_validation

        # Example callback running after each torch model forward pass in validation.
        >>> @on_forward_validation
        ... def print_callback(state):
        ...     print('Evaluated validation batch.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Evaluated validation batch.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_forward_validation` calling func
    """
    return bind_to(Callback.on_forward_validation)(func)


def on_criterion_validation(func):
    """ The :func:`on_criterion_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_criterion_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_criterion_validation

        # Example callback running after each criterion evaluation in validation.
        >>> @on_criterion_validation
        ... def print_callback(state):
        ...     print('Current val loss {}.'.format(state[torchbearer.LOSS].item()))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Current val loss 0.0.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_criterion_validation` calling func
    """
    return bind_to(Callback.on_criterion_validation)(func)


def on_end_validation(func):
    """ The :func:`on_end_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_end_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_end_validation

        # Example callback running at the end of each validation pass.
        >>> @on_end_validation
        ... def print_callback(state):
        ...     print('Finished validating.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Finished validating.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_end_validation` calling func
    """
    return bind_to(Callback.on_end_validation)(func)


def on_step_validation(func):
    """ The :func:`on_step_validation` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_step_validation`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_step_validation

        # Example callback running at the end of each validation step.
        >>> @on_step_validation
        ... def print_callback(state):
        ...     print('Validation step {}.'.format(state[torchbearer.BATCH]))

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).for_val_steps(1).run()
        Validation step 0.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_step_validation` calling func
    """
    return bind_to(Callback.on_step_validation)(func)


def on_checkpoint(func):
    """ The :func:`on_checkpoint` decorator is used to initialise a :class:`.Callback` with :meth:`~.Callback.on_checkpoint`
    calling the decorated function

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import on_checkpoint

        # Example callback running at checkpoint time.
        >>> @on_checkpoint
        ... def print_callback(state):
        ...     print('Checkpointing.')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        Checkpointing.

    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback with :meth:`~.Callback.on_checkpoint` calling func
    """
    return bind_to(Callback.on_checkpoint)(func)


def add_to_loss(func):
    """ The :func:`add_to_loss` decorator is used to initialise a :class:`.Callback` with the value returned from func
    being added to the loss

    Example: ::

        >>> import torch
        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import add_to_loss

        # Example callback to add a quantity to the loss each step.
        >>> @add_to_loss
        ... def loss_callback(state):
        ...     return torch.Tensor([1.125])

        >>> trial = Trial(None, callbacks=[loss_callback], metrics=['loss']).for_steps(1).run()
        >>> print(trial[0][1]['loss'])
        1.125


    Args:
        func (function): The function(state) to *decorate*

    Returns:
        Callback: Initialised callback which adds the returned value from func to the loss
    """
    @on_criterion
    @on_criterion_validation
    def add_to_loss_callback(state):
        state[torchbearer.LOSS] = state[torchbearer.LOSS] + func(state)

    return add_to_loss_callback


def once(fcn):
    """
    Decorator to fire a callback once in the lifetime of the callback. If the callback is a class method, each
    instance of the class will fire only once. For functions, only the first instance will fire (even if more than
    one function is present in the callback list).

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import once, on_step_training

        # Example callback to be called exactly once on the very first training step
        >>> @once
        ... @on_step_training
        ... def print_callback(state):
        ...     print('This happens once ever')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        This happens once ever

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    def _once(self, _):
        try:
            return not self.__done__
        except AttributeError:
            self.__done__ = True
            return True

    return only_if(_once)(fcn)


def once_per_epoch(fcn):
    """Decorator to fire a callback once (on the first call) in any given epoch. If the callback is a class method, each
    instance of the class will fire once per epoch. For functions, only the first instance will fire (even if more than
    one function is present in the callback list).

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import once_per_epoch, on_step_training

        # Example callback to be called exactly once per epoch, on the first training step
        >>> @once_per_epoch
        ... @on_step_training
        ... def print_callback(state):
        ...     print('This happens once per epoch')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(1).run()
        This happens once per epoch

    .. note::
        The decorated callback may exhibit unusual behaviour if it is reused

    Args:
        fcn (function): the `torchbearer callback` function to decorate.

    Returns:
        the decorator
    """
    def ope(self, state):
        try:
            if state[torchbearer.EPOCH] != self.__last_epoch__:
                self.__last_epoch__ = state[torchbearer.EPOCH]
                return True
            return False
        except AttributeError:
            self.__last_epoch__ = state[torchbearer.EPOCH]
            return True

    return only_if(ope)(fcn)


def only_if(condition_expr):
    """
    Decorator to fire a callback only if the given conditional expression function returns True. The conditional
    expression can be a function of state or self and state. If the decorated function is not a class method (i.e. it
    does not take state) the decorated function will be passed instead. This enables the storing of temporary variables.

    Example: ::

        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import only_if, on_step_training

        # Example callback to be called only when the given condition is true on each training step
        >>> @only_if(lambda state: state[torchbearer.BATCH] == 100)
        ... @on_step_training
        ... def print_callback(state):
        ...     print('This is the 100th batch')

        >>> trial = Trial(None, callbacks=[print_callback]).for_steps(101).run()
        This is the 100th batch

    Args:
        condition_expr (function(self, state) or function(self)): a function/lambda which takes state and optionally\
                        self that must evaluate to true for the decorated `torchbearer callback` to be called. The\
                        `state` object passed to the callback will be passed as an argument to the condition function.

    Returns:
        the decorator
    """
    def condition_decorator(fcn):
        if isinstance(fcn, LambdaCallback):
            fcn.func = condition_decorator(fcn.func)
            return fcn
        else:
            count = count_args(fcn)
            if count == 2 and not hasattr(fcn, '__self__'):  # Assume Class method
                def decfcn(o, state):
                    try:
                        res = condition_expr(o, state)
                    except TypeError:
                        res = condition_expr(state)
                    if res:
                        return fcn(o, state)
            else:  # Assume function of state
                def id_fcn(state):
                    return fcn(state)  # Hack to allow setting attributes of bound methods

                def decfcn(state):
                    try:
                        res = condition_expr(id_fcn, state)
                    except TypeError:
                        res = condition_expr(state)
                    if res:
                        return id_fcn(state)
            return decfcn
    return condition_decorator



import warnings

from unittest import TestCase
from mock import MagicMock, Mock

import torchbearer
from torchbearer.callbacks import CallbackList, Tqdm, TensorBoard


class TestCallbackList(TestCase):
    def __init__(self, methodName='runTest'):
        super(TestCallbackList, self).__init__(methodName)
        self.callback_1 = MagicMock(spec=torchbearer.callbacks.printer.Tqdm())
        self.callback_2 = MagicMock(spec=torchbearer.callbacks.tensor_board.TensorBoard())
        callbacks = [self.callback_1, self.callback_2]
        self.list = CallbackList(callbacks)

    def test_state_dict(self):
        self.callback_1.state_dict = Mock(return_value='test_1')
        self.callback_2.state_dict = Mock(return_value='test_2')

        state = self.list.state_dict()

        self.assertEqual(self.callback_1.state_dict.call_count, 1)
        self.assertEqual(self.callback_2.state_dict.call_count, 1)
        self.assertEqual(state[CallbackList.CALLBACK_STATES][0], 'test_1')
        self.assertEqual(state[CallbackList.CALLBACK_STATES][1], 'test_2')
        self.assertEqual(state[CallbackList.CALLBACK_TYPES][0], Tqdm().__class__)
        self.assertEqual(state[CallbackList.CALLBACK_TYPES][1], TensorBoard().__class__)

    def test_load_state_dict(self):
        self.callback_1.load_state_dict = Mock(return_value='test_1')
        self.callback_2.load_state_dict = Mock(return_value='test_2')

        self.callback_1.state_dict = Mock(return_value='test_1')
        self.callback_2.state_dict = Mock(return_value='test_2')

        state = self.list.state_dict()
        self.list.load_state_dict(state)

        self.callback_1.load_state_dict.assert_called_once_with('test_1')
        self.callback_2.load_state_dict.assert_called_once_with('test_2')

        state = self.list.state_dict()
        state[CallbackList.CALLBACK_TYPES] = list(reversed(state[CallbackList.CALLBACK_TYPES]))

        with warnings.catch_warnings(record=True) as w:
            self.list.load_state_dict(state)
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))
            self.assertTrue('Callback classes did not match, expected: [\'TensorBoard\', \'Tqdm\']' in str(w[-1].message))

    def test_for_list(self):
        self.list.on_start({})
        self.assertTrue(self.callback_1.method_calls[0][0] == 'on_start')
        self.assertTrue(self.callback_2.method_calls[0][0] == 'on_start')

    def test_list_in_list(self):
        callback = 'test'
        clist = CallbackList([callback])
        clist2 = CallbackList([clist])
        self.assertTrue(clist2.callback_list[0] == 'test')

    def test_iter_copy(self):
        callback = 'test'
        clist = CallbackList([callback])
        cpy = clist.__copy__()
        self.assertTrue(cpy.callback_list[0] == 'test')
        self.assertTrue(cpy is not clist)
        cpy = clist.copy()
        self.assertTrue(cpy.callback_list[0] == 'test')
        self.assertTrue(cpy is not clist)
        for cback in clist:
            self.assertTrue(cback == 'test')



from unittest import TestCase
from mock import patch, call
import torch

import torchbearer
from torchbearer.callbacks import SamplePairing


class TestSamplePairing(TestCase):
    @patch('torchbearer.callbacks.sample_pairing.torch')
    def test_on_sample(self, mock_torch):
        mock_torch.randperm.return_value = torch.Tensor([3, 2, 1, 0]).long()
        callback = SamplePairing(policy=SamplePairing.default_policy(0, 10, 10, 10))

        state = {torchbearer.INPUT: torch.Tensor([0.25, 0.5, 0.75, 1]), torchbearer.EPOCH: 0}
        callback.on_sample(state)
        self.assertTrue(((state[torchbearer.INPUT] - 0.625).abs() < 0.0001).all())

    def test_default_policy(self):
        policy = SamplePairing.default_policy(10, 80, 3, 2)

        state = {torchbearer.EPOCH: 0}
        self.assertFalse(policy(state))
        state = {torchbearer.EPOCH: 10}
        self.assertTrue(policy(state))

        for i in range(2):
            state = {torchbearer.EPOCH: 11 + i}
            policy(state)
        state = {torchbearer.EPOCH: 14}
        self.assertFalse(policy(state))
        state = {torchbearer.EPOCH: 15}
        self.assertFalse(policy(state))
        state = {torchbearer.EPOCH: 16}
        self.assertTrue(policy(state))



import sys
import os

import torchbearer
from torchbearer.callbacks import Callback


class no_print:
    def __init__(self):
        pass

    def __enter__(self):
        self.stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')
        return self

    def __exit__(self, *exc):
        sys.stdout = self.stdout
        return False


class LiveLossPlot(Callback):
    """
    Callback to write metrics to `LiveLossPlot <https://github.com/stared/livelossplot>`_, a library for visualisation in notebooks

    Example: ::

        >>> import torch.nn
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import LiveLossPlot

        # Example Trial which clips all model gradients norms at 2 under the L1 norm.
        >>> model = torch.nn.Linear(1,1)
        >>> live_loss_plot = LiveLossPlot()
        >>> trial = Trial(model, callbacks=[live_loss_plot], metrics=['acc'])

    Args:
        on_batch (bool): If True, batch metrics will be logged. Else batch metrics will not be logged
        batch_step_size (int): The number of batches between logging metrics
        on_epoch (bool): If True, epoch metrics will be logged every epoch. Else epoch metrics will not be logged
        draw_once (bool): If True, draw the plot only at the end of training. Else draw every time metrics are logged
        kwargs: Keyword arguments for livelossplot.PlotLosses

    State Requirements:
        - :attr:`torchbearer.state.METRICS`: Metrics should be a dict containing the metrics to be plotted
        - :attr:`torchbearer.state.BATCH`: Batch should be the current batch or iteration number in the epoch
    """
    def __init__(self, on_batch=False, batch_step_size=10, on_epoch=True, draw_once=False, **kwargs):
        super(LiveLossPlot, self).__init__()
        self._kwargs = kwargs

        self.on_batch = on_batch
        self.on_epoch = on_epoch
        self.draw_once = draw_once
        self.batch_step_size = batch_step_size

        if on_batch:
            self.on_step_training = self._on_step_training

        if on_epoch:
            self.on_end_epoch = self._on_end_epoch

    def on_start(self, state):
        from livelossplot import PlotLosses
        self.plt = PlotLosses(**self._kwargs)
        self.batch_plt = PlotLosses(**self._kwargs)

    def _on_step_training(self, state):
        self.batch_plt.update({k: state[torchbearer.METRICS][k] for k in state[torchbearer.METRICS]})
        if state[torchbearer.BATCH] % self.batch_step_size == 0 and not self.draw_once:
            with no_print():
                self.batch_plt.draw()

    def _on_end_epoch(self, state):
        self.plt.update({k: state[torchbearer.METRICS][k] for k in state[torchbearer.METRICS]})
        if not self.draw_once:
            with no_print():
                self.plt.draw()

    def on_end(self, state):
        if self.draw_once:
            with no_print():
                self.batch_plt.draw()
                self.plt.draw()



import unittest

import torch
from torch.nn import Module, Linear
import torch.nn.init as init

import torchbearer


class Net(Module):
    def __init__(self, x):
        super(Net, self).__init__()
        self.pars = torch.nn.Parameter(x)

    def f(self):
        """
        function to be minimised:
        f(x) = (x[0]-5)^2 + x[1]^2 + (x[2]-1)^2
        Solution:
        x = [5,0,1]
        """
        out = torch.zeros_like(self.pars)
        out[0] = self.pars[0]-5
        out[1] = self.pars[1]
        out[2] = self.pars[2]-1
        return torch.sum(out**2)

    def forward(self, _):
        return self.f()


class NetWithState(Net):
    def forward(self, _, state=None):
        if state is None:
            raise ValueError
        return super(NetWithState, self).forward(_)


def loss(y_pred, y_true):
    return y_pred


class TestEndToEnd(unittest.TestCase):
    def test_basic_opt(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        training_steps = 1000

        model = NetWithState(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        trial = torchbearer.Trial(model, optim, loss).for_train_steps(training_steps).for_val_steps(1).for_test_steps(1)
        trial.run()
        trial.predict()
        trial.evaluate()

        self.assertAlmostEqual(model.pars[0].item(), 5.0, places=4)
        self.assertAlmostEqual(model.pars[1].item(), 0.0, places=4)
        self.assertAlmostEqual(model.pars[2].item(), 1.0, places=4)

    def test_zero_model(self):
        model = Linear(3, 1)
        init.constant_(model.weight, 0)
        init.constant_(model.bias, 0)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        trial = torchbearer.Trial(model, optim, loss)
        trial.with_test_data(torch.rand(10, 3), batch_size=3)
        preds = trial.predict()

        for i in range(len(preds)):
            self.assertAlmostEqual(preds[i], 0)

    def test_basic_checkpoint(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        training_steps = 500

        model = Net(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        trial = torchbearer.Trial(model, optim, loss, callbacks=[torchbearer.callbacks.MostRecent(filepath='test.pt')]).for_train_steps(training_steps).for_val_steps(1)
        trial.run(2)  # Simulate 2 'epochs'

        # Reload
        p = torch.tensor([2.0, 1.0, 10.0])
        model = Net(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)

        trial = torchbearer.Trial(model, optim, loss, callbacks=[torchbearer.callbacks.MostRecent(filepath='test.pt')]).for_train_steps(training_steps)
        trial.load_state_dict(torch.load('test.pt'))
        self.assertEqual(len(trial.state[torchbearer.HISTORY]), 2)
        self.assertAlmostEqual(model.pars[0].item(), 5.0, places=4)
        self.assertAlmostEqual(model.pars[1].item(), 0.0, places=4)
        self.assertAlmostEqual(model.pars[2].item(), 1.0, places=4)

        import os
        os.remove('test.pt')

    def test_with_loader(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        training_steps = 2

        model = Net(p)
        optim = torch.optim.SGD(model.parameters(), lr=0.01)
        test_var = {'loaded': False}

        def custom_loader(state):
            state[torchbearer.X], state[torchbearer.Y_TRUE] = None, None
            test_var['loaded'] = True

        trial = torchbearer.Trial(model, optim, loss, callbacks=[torchbearer.callbacks.MostRecent(filepath='test.pt')]).for_train_steps(training_steps).for_val_steps(1)
        trial.with_loader(custom_loader)
        self.assertTrue(not test_var['loaded'])
        trial.run(1)
        self.assertTrue(test_var['loaded'])

        import os
        os.remove('test.pt')

    def test_only_model(self):
        p = torch.tensor([2.0, 1.0, 10.0])
        model = Net(p)
        trial = torchbearer.Trial(model)
        self.assertListEqual(trial.run(), [])

    def test_no_model(self):
        trial = torchbearer.Trial(None)
        trial.run()
        self.assertTrue(torchbearer.trial.MockModel()(torch.rand(1)) is None)

    def test_no_train_steps(self):
        trial = torchbearer.Trial(None)
        trial.for_val_steps(10)
        trial.run()


import torch
import torch.nn as nn
import torch.optim as optim
import torchbearer
from torchbearer.callbacks.decorators import once_per_epoch
from . import imaging

_inside_cnns = """
@article{simonyan2013deep,
  title={Deep inside convolutional networks: Visualising image classification models and saliency maps},
  author={Simonyan, Karen and Vedaldi, Andrea and Zisserman, Andrew},
  journal={arXiv preprint arXiv:1312.6034},
  year={2013}
}
"""

RANDOM = -10
""" Flag that when passed as the target chosses a random target"""


class _CAMWrapper(nn.Module):
    def __init__(self, input_size, base_model, transform=None):
        super(_CAMWrapper, self).__init__()
        self.base_model = base_model
        input_image = torch.zeros(input_size)

        self.input_image = nn.Parameter(input_image, requires_grad=True)

        self.transform = (lambda x: x) if transform is None else transform

    def forward(self, _, state):
        try:
            return self.base_model(self.transform(self.input_image.sigmoid()).unsqueeze(0), state)
        except TypeError:
            return self.base_model(self.transform(self.input_image.sigmoid()).unsqueeze(0))


def _cam_loss(key, targets_hot, decay):
    def loss(state):
        img = state[torchbearer.MODEL].input_image
        return - torch.masked_select(state[key], targets_hot).sum() + decay * img.pow(2).sum()
    return loss


@torchbearer.cite(_inside_cnns)
class ClassAppearanceModel(imaging.ImagingCallback):
    """The :class:`.ClassAppearanceModel` callback implements Figure 1 from
    `Deep Inside Convolutional Networks: Visualising Image Classification Models and Saliency Maps <https://arxiv.org/abs/1312.6034>`_.
    This is a simple gradient ascent on an image (initialised to zero) with a sum-squares regularizer. Internally this
    creates a new :class:`.Trial` instance which then performs the optimization.

    Args:
        nclasses (int): The number of output classes
        input_size (tuple): The size to use for the input image
        optimizer_factory: A function of parameters which returns an optimizer to use
        logit_key (StateKey): :class:`.StateKey` storing the class logits
        target (int): Target class for the optimisation or RANDOM
        steps (int): Number of optimisation steps to take
        decay (float): Lambda for the L2 decay on the image
        verbose (int): Verbosity level to pass to the internal :class:`.Trial` instance
        transform (callable, optional): A function/transform that  takes in a Tensor and returns a transformed version.
            This will be applied to the image before it is sent to output

    """
    def __init__(self, nclasses, input_size, optimizer_factory=lambda params: optim.Adam(params, lr=0.5), steps=256,
                 logit_key=torchbearer.PREDICTION, target=RANDOM, decay=0.01, verbose=0, in_transform=None, transform=None):
        super(ClassAppearanceModel, self).__init__(transform=transform)

        self.nclasses = nclasses
        self.input_size = input_size
        self.optimizer_factory = optimizer_factory
        self.logit_key = logit_key
        self.target = target
        self.steps = steps
        self.decay = decay
        self.verbose = verbose
        self.in_transform = in_transform

        self._target_keys = []

    def target_to_key(self, key):
        self._target_keys.append(key)
        return self

    def _targets_hot(self, state):
        targets = torch.randint(high=self.nclasses, size=(1, 1)).long().to(state[torchbearer.DEVICE])
        if self.target is not RANDOM:
            targets[0][0] = self.target
        for key in self._target_keys:
            state[key] = targets
        targets_hot = torch.zeros(1, self.nclasses).to(state[torchbearer.DEVICE])
        targets_hot.scatter_(1, targets, 1)
        targets_hot = targets_hot.ge(0.5)
        return targets_hot

    @torchbearer.enable_grad()
    @once_per_epoch
    def on_batch(self, state):
        training = state[torchbearer.MODEL].training
        state[torchbearer.MODEL].eval()

        targets_hot = self._targets_hot(state)

        key = self.logit_key

        @torchbearer.callbacks.on_sample
        def make_eval(_):
            state[torchbearer.MODEL].eval()

        model = _CAMWrapper(self.input_size, state[torchbearer.MODEL], transform=self.in_transform)
        trial = torchbearer.Trial(model, self.optimizer_factory(filter(lambda p: p.requires_grad, [model.input_image])),
                                  _cam_loss(key, targets_hot, self.decay), callbacks=[make_eval])
        trial.for_train_steps(self.steps).to(state[torchbearer.DEVICE], state[torchbearer.DATA_TYPE])
        trial.run(verbose=self.verbose)

        if training:
            state[torchbearer.MODEL].train()

        return model.input_image.sigmoid()



import unittest

import torchbearer.callbacks as callbacks
import torchbearer


class TestDecorators(unittest.TestCase):

    def test_multi(self):
        def example(state):
            return state
        state = 'test'
        c = callbacks.on_backward(callbacks.on_sample(callbacks.on_start(example)))
        self.assertTrue(c.on_backward(state) == state)
        self.assertTrue(c.on_sample(state) == state)
        self.assertTrue(c.on_start(state) == state)

    def test_target(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.bind_to(callbacks.on_start)(example).on_start(state) == state)

    def test_on_init(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_init(example).on_init(state) == state)

    def test_on_start(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_start(example).on_start(state) == state)

    def test_on_start_epoch(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_start_epoch(example).on_start_epoch(state) == state)

    def test_on_start_training(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_start_training(example).on_start_training(state) == state)

    def test_on_sample(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_sample(example).on_sample(state) == state)

    def test_on_forward(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_forward(example).on_forward(state) == state)

    def test_on_criterion(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_criterion(example).on_criterion(state) == state)

    def test_on_backward(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_backward(example).on_backward(state) == state)

    def test_on_step_training(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_step_training(example).on_step_training(state) == state)

    def test_on_end_training(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_end_training(example).on_end_training(state) == state)

    def test_on_end_epoch(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_end_epoch(example).on_end_epoch(state) == state)

    def test_on_end(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_end(example).on_end(state) == state)

    def test_on_start_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_start_validation(example).on_start_validation(state) == state)

    def test_on_sample_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_sample_validation(example).on_sample_validation(state) == state)

    def test_on_forward_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_forward_validation(example).on_forward_validation(state) == state)

    def test_on_criterion_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_criterion_validation(example).on_criterion_validation(state) == state)

    def test_on_end_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_end_validation(example).on_end_validation(state) == state)

    def test_on_step_validation(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_step_validation(example).on_step_validation(state) == state)

    def test_on_checkpoint(self):
        def example(state):
            return state
        state = 'test'
        self.assertTrue(callbacks.on_checkpoint(example).on_checkpoint(state) == state)

    def test_add_to_loss(self):
        def example(state):
            return 1
        state = {'test': 'test', torchbearer.LOSS: 0}
        callbacks.add_to_loss(example).on_criterion(state)
        self.assertTrue(state[torchbearer.LOSS] == 1)
        callbacks.add_to_loss(example).on_criterion_validation(state)
        self.assertTrue(state[torchbearer.LOSS] == 2)

    def test_once(self):
        class Example(callbacks.Callback):
            @callbacks.once
            def on_step_validation(self, state):
                state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = Example()

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb2 = Example()

        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

    def test_once_per_epoch(self):
        class Example(callbacks.Callback):
            @callbacks.once_per_epoch
            def on_step_validation(self, state):
                state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = Example()

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb2 = Example()

        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb2.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

    def test_only_if(self):
        class Example(callbacks.Callback):
            @callbacks.only_if(lambda s: s[torchbearer.EPOCH] % 2 == 0)
            def on_step_validation(self, state):
                state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = Example()

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 3)

    def test_once_lambda(self):
        @callbacks.once
        @callbacks.on_step_validation
        def callback_func(state):
            state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = callback_func

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

    def test_once_per_epoch_lambda(self):
        @callbacks.once_per_epoch
        @callbacks.on_step_validation
        def callback_func(state):
            state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = callback_func

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

    def test_only_if_lambda(self):
        @callbacks.only_if(lambda s: s[torchbearer.EPOCH] % 2 == 0)
        @callbacks.on_step_validation
        def callback_func(state):
            state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = callback_func

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 3)

    def test_lambda_only_if(self):
        @callbacks.on_step_validation
        @callbacks.only_if(lambda s: s[torchbearer.EPOCH] % 2 == 0)
        def callback_func(state):
            state['value'] += 1

        state = {torchbearer.EPOCH: 0, 'value': 0}

        cb = callback_func

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 1)

        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 2)

        state[torchbearer.EPOCH] += 1
        cb.on_step_validation(state)
        self.assertTrue(state['value'] == 3)


from unittest import TestCase
from mock import patch, call, MagicMock
import torch

import torchbearer
from torchbearer.callbacks import Mixup, MixupAcc


class TestMixupInputs(TestCase):
    def test_input(self):
        mixup = Mixup()
        X = torch.Tensor([[1., 2., 3.], [6., 7., 8.]])
        Y_true = torch.Tensor([0., 1.])

        state = {
            torchbearer.X : X,
            torchbearer.Y_TRUE : Y_true
        }

        mixup.on_sample(state)
        self.assertTrue((torch.eq(state[torchbearer.X], X * state[torchbearer.MIXUP_LAMBDA] + X[state[torchbearer.MIXUP_PERMUTATION], :] * (1-state[torchbearer.MIXUP_LAMBDA]))).all())

    def test_alpha(self):
        mixup = Mixup(-0.1)
        X = torch.Tensor([[1., 2., 3.], [6., 7., 8.]])
        Y_true = torch.Tensor([0., 1.])

        state = {
            torchbearer.X : X,
            torchbearer.Y_TRUE : Y_true
        }

        mixup.on_sample(state)
        self.assertTrue(state[torchbearer.MIXUP_LAMBDA] == 1.0)

    def test_fixed_lambda(self):
        mixup = Mixup(-0.1, 0.3)
        X = torch.Tensor([[1., 2., 3.], [6., 7., 8.]])
        Y_true = torch.Tensor([0., 1.])
        lam = 0.3

        state = {
            torchbearer.X : X,
            torchbearer.Y_TRUE : Y_true,
            torchbearer.MIXUP_LAMBDA : lam
        }

        mixup.on_sample(state)
        self.assertTrue(state[torchbearer.X][0][1] == 0.3 * 2 + 0.7 * 2 or state[torchbearer.X][0][1] == 0.3 * 2 + 0.7 * 7)

    def test_target(self):
        mixup = Mixup(-0.1)
        X = torch.Tensor([[1., 2., 3.], [6., 7., 8.]])
        Y_true = torch.Tensor([0., 1.])

        state = {
            torchbearer.X : X,
            torchbearer.Y_TRUE : Y_true
        }

        mixup.on_sample(state)

        self.assertTrue((state[torchbearer.Y_TRUE][0] == torch.Tensor([0., 1.])).all())
        self.assertTrue((state[torchbearer.Y_TRUE][1] == torch.Tensor([0., 1.])).all() or (state[torchbearer.Y_TRUE][1] == torch.Tensor([1., 0.])).all())

    @patch('torchbearer.callbacks.mixup.F.cross_entropy')
    def test_loss(self, mock_cross_entropy):
        mock_cross_entropy.return_value = 1.0

        loss = Mixup.mixup_loss
        res = loss({torchbearer.DATA: torchbearer.TRAIN_DATA, torchbearer.Y_PRED: 'test1', torchbearer.Y_TRUE: ('target1', 'target2'), torchbearer.MIXUP_LAMBDA: 0.7})

        mock_cross_entropy.assert_has_calls([call('test1', 'target1'), call('test1', 'target2')])
        self.assertTrue(res == 1.0)

class TestMixupAcc(TestCase):
    def setUp(self):
        self.lam = 0.8

        self._state = {
            torchbearer.Y_TRUE: (torch.LongTensor([0, 1, 2, 2, 1]), torch.LongTensor([1, 2, 1, 0, 1])),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1],  # Correct
                [0.1, 0.9, 0.1],  # Correct
                [0.1, 0.1, 0.9],  # Correct
                [0.9, 0.1, 0.1],  # Incorrect
                [0.9, 0.1, 0.1],  # Incorrect
            ]),
            torchbearer.MIXUP_LAMBDA: self.lam,
        }
        self._targets = [0.8, 0.8, 0.8, 0.2, 0]
        self._true_targets = [1, 1, 1, 0, 0]
        self._metric = MixupAcc().root  # Get root node of Tree for testing

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process_train(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_val_process(self):
        self._metric.train()
        _state = self._state.copy()
        _state[torchbearer.Y_TRUE] = _state[torchbearer.Y_TRUE][0]
        result = self._metric.process_validate(_state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._true_targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_reset(self):
        self._metric.cat_acc = MagicMock()
        self._metric.reset({})
        self.assertTrue(self._metric.cat_acc.reset.call_count == 1)


import unittest
from mock import Mock, MagicMock, create_autospec, patch

import torch

import torchbearer
from torchbearer.bases import Metric, Callback, base_closure
from torchbearer.metrics.primitives import CategoricalAccuracy


class TestMetric(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: torch.LongTensor([0, 1, 2, 2, 1]),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1], # Correct
                [0.1, 0.9, 0.1], # Correct
                [0.1, 0.1, 0.9], # Correct
                [0.9, 0.1, 0.1], # Incorrect
                [0.9, 0.1, 0.1], # Incorrect
            ])
        }
        self._state[torchbearer.Y_PRED].requires_grad = True
        self._targets = [1, 1, 1, 0, 0]
        self._metric = CategoricalAccuracy().root

    def test_requires_grad(self):
        result = self._metric.process(self._state)
        self.assertTrue(self._state[torchbearer.Y_PRED].requires_grad is True)
        self.assertTrue(result.requires_grad is False)

    def test_empty_methods(self):
        metric = Metric(name='test')
        self.assertTrue(metric.process() is None)
        self.assertTrue(metric.process_final() is None)


class TestCallback(unittest.TestCase):
    def test_state_dict(self):
        callback = Callback()

        self.assertEqual(callback.state_dict(), {})
        self.assertEqual(callback.load_state_dict({}), callback)

    def test_str(self):
        callback = Callback()
        self.assertEqual(str(callback).strip(), "torchbearer.bases.Callback")

    def test_empty_methods(self):
        callback = Callback()

        self.assertIsNone(callback.on_init({}))
        self.assertIsNone(callback.on_start({}))
        self.assertIsNone(callback.on_start_epoch({}))
        self.assertIsNone(callback.on_start_training({}))
        self.assertIsNone(callback.on_sample({}))
        self.assertIsNone(callback.on_forward({}))
        self.assertIsNone(callback.on_criterion({}))
        self.assertIsNone(callback.on_backward({}))
        self.assertIsNone(callback.on_step_training({}))
        self.assertIsNone(callback.on_end_training({}))
        self.assertIsNone(callback.on_end_epoch({}))
        self.assertIsNone(callback.on_checkpoint({}))
        self.assertIsNone(callback.on_end({}))
        self.assertIsNone(callback.on_start_validation({}))
        self.assertIsNone(callback.on_sample_validation({}))
        self.assertIsNone(callback.on_forward_validation({}))
        self.assertIsNone(callback.on_end_validation({}))
        self.assertIsNone(callback.on_step_validation({}))
        self.assertIsNone(callback.on_criterion_validation({}))


class TestBaseCrit(unittest.TestCase):

    def state_model_with_e(self, e):
        def model(x, state):
            raise e
            return x
        return model

    def stateless_model_with_e(self, e):
        def model(x):
            raise e
            return x
        return model

    def optional_model_with_e(self, e):
        def model(x, state=None):
            raise e
            return x
        return model

    def test_opt(self):
        opt = Mock()
        opt.zero_grad = Mock()
        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(opt.zero_grad.call_count == 1)

    def test_forward_x(self):
        opt = Mock()

        def model_forward(x):
            return None
        model = create_autospec(model_forward)

        x = 'test'

        state = {torchbearer.X: x, torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(model.call_args[0][0] == x)

    def test_forward_multiple_x(self):
        opt = Mock()

        def model_forward(x1, x2):
            return None
        model = create_autospec(model_forward)

        x1 = 'test1'
        x2 = 'test2'

        state = {torchbearer.X: [x1, x2], torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(model.call_args[0][0] == x1)
        self.assertTrue(model.call_args[0][1] == x2)

    def test_forward_state(self):
        opt = Mock()

        def model_forward(x, state):
            return None
        model = create_autospec(model_forward)

        x = 'test'

        state = {torchbearer.X: x, torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(model.call_args[0][0] == x)
        self.assertDictEqual(model.call_args[1]['state'], state)

    def test_forward_multiple_x_and_state(self):
        opt = Mock()

        def model_forward(x1, x2, state):
            return None
        model = create_autospec(model_forward)

        x1 = 'test1'
        x2 = 'test2'

        state = {torchbearer.X: [x1, x2], torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(model.call_args[0][0] == x1)
        self.assertTrue(model.call_args[0][1] == x2)
        self.assertDictEqual(model.call_args[1]['state'], state)

    def test_loss_no_state(self):
        opt = Mock()

        y_pred = 'yp'
        y_true = 'yt'
        def loss_sig(y_pred, y_true):
            return None
        crit = create_autospec(loss_sig)

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: y_pred, torchbearer.Y_TRUE: y_true,
                 torchbearer.CRITERION: crit, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(crit.call_args[0] == (y_pred, y_true))

    def test_loss_multiple_output_no_state(self):
        opt = Mock()

        y_pred1 = 'yp1'
        y_pred2 = 'yp2'
        y_true1 = 'yt1'
        y_true2 = 'yt2'
        def loss_sig(y_pred1, y_pred2, y_true1, y_true2):
            return None
        crit = create_autospec(loss_sig)

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: (y_pred1, y_pred2), torchbearer.Y_TRUE: (y_true1, y_true2),
                 torchbearer.CRITERION: crit, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(crit.call_args[0] == (y_pred1, y_pred2, y_true1, y_true2))

    def test_loss_state(self):
        opt = Mock()

        y_pred = 'yp'
        y_true = 'yt'
        def loss_sig(state):
            return None
        crit = create_autospec(loss_sig)

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: y_pred, torchbearer.Y_TRUE: y_true,
                 torchbearer.CRITERION: crit, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertDictEqual(crit.call_args[0][0], state)

    def test_backward(self):
        opt = Mock()
        loss = Mock()
        loss.backward = Mock()

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda state: loss, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(loss.backward.call_count == 1)

    def test_callback_list(self):
        opt = Mock()
        callback_list = Mock()
        callback_list.on_forward = Mock()
        callback_list.on_criterion = Mock()
        callback_list.on_backward = Mock()

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda state: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: callback_list, torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)
        closure(state)
        self.assertTrue(callback_list.on_forward.call_count == 1)
        self.assertTrue(callback_list.on_criterion.call_count == 1)
        self.assertTrue(callback_list.on_backward.call_count == 1)


    def test_stateless_type_error(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.stateless_model_with_e(TypeError('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 2)

    def test_stateless_exception(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.stateless_model_with_e(Exception('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 1)
        self.assertTrue('test' in context.exception.args[0][0].args)

    def test_exception_exception(self):

        state = {torchbearer.X: None, torchbearer.MODEL: self.optional_model_with_e(Exception('test')),
                 torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 2)
        self.assertTrue('test' in context.exception.args[0][0].args)
        self.assertTrue('test' in context.exception.args[0][1].args)


    def test_state_exception(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.state_model_with_e(Exception('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 1)
        self.assertTrue('test' in context.exception.args[0][0].args)

    def test_state_type_error(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.state_model_with_e(TypeError('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = base_closure(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, torchbearer.Y_TRUE,
                               torchbearer.CRITERION, torchbearer.LOSS, torchbearer.OPTIMIZER)

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 2)


class TestApexCrit(unittest.TestCase):
    def setUp(self):
        super(TestApexCrit, self).setUp()

        import types
        import sys

        module_name = 'apex'
        bogus_module = types.ModuleType(module_name)
        self.old_module = sys.modules[module_name] if module_name in sys.modules else None
        sys.modules[module_name] = bogus_module
        self.mock_amp = MagicMock(name=module_name + '.amp')
        bogus_module.amp = self.mock_amp

        from torchbearer.bases import apex_closure
        self.closure = apex_closure()

    def tearDown(self):
        super(TestApexCrit, self).tearDown()
        if self.old_module is not None:
            import sys
            sys.modules['apex'] = self.old_module

    def state_model_with_e(self, e):
        def model(x, state):
            raise e
            return x
        return model

    def stateless_model_with_e(self, e):
        def model(x):
            raise e
            return x
        return model

    def test_opt(self):
        opt = Mock()
        opt.zero_grad = Mock()
        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue(opt.zero_grad.call_count == 1)

    def test_forward_x(self):
        opt = Mock()

        def model_forward(x):
            return None
        model = create_autospec(model_forward)

        x = 'test'

        state = {torchbearer.X: x, torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue(model.call_args[0][0] == x)

    def test_forward_state(self):
        opt = Mock()

        def model_forward(x, state):
            return None
        model = create_autospec(model_forward)

        x = 'test'

        state = {torchbearer.X: x, torchbearer.MODEL: model, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda x,y: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue(model.call_args[0][0] == x)
        self.assertDictEqual(model.call_args[1]['state'], state)

    def test_loss_no_state(self):
        opt = Mock()

        y_pred = 'yp'
        y_true = 'yt'
        def loss_sig(y_pred, y_true):
            return None
        crit = create_autospec(loss_sig)

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: y_pred, torchbearer.Y_TRUE: y_true,
                 torchbearer.CRITERION: crit, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue(crit.call_args[0] == (y_pred, y_true))

    def test_loss_state(self):
        opt = Mock()

        y_pred = 'yp'
        y_true = 'yt'
        def loss_sig(state):
            return None
        crit = create_autospec(loss_sig)

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: y_pred, torchbearer.Y_TRUE: y_true,
                 torchbearer.CRITERION: crit, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertDictEqual(crit.call_args[0][0], state)

    def test_backward(self):
        opt = Mock()
        loss = Mock()
        loss.backward = Mock()

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda state: loss, torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: MagicMock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue('backward' in self.mock_amp.scale_loss.mock_calls[2][0])

    def test_callback_list(self):
        opt = Mock()
        callback_list = Mock()
        callback_list.on_forward = Mock()
        callback_list.on_criterion = Mock()
        callback_list.on_backward = Mock()

        state = {torchbearer.X: None, torchbearer.MODEL: lambda x: None, torchbearer.Y_TRUE: None,
                 torchbearer.CRITERION: lambda state: MagicMock(), torchbearer.LOSS: None, torchbearer.OPTIMIZER: opt,
                 torchbearer.CALLBACK_LIST: callback_list, torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure
        closure(state)
        self.assertTrue(callback_list.on_forward.call_count == 1)
        self.assertTrue(callback_list.on_criterion.call_count == 1)
        self.assertTrue(callback_list.on_backward.call_count == 1)


    def test_stateless_type_error(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.stateless_model_with_e(TypeError('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 2)

    def test_stateless_exception(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.stateless_model_with_e(Exception('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 1)
        self.assertTrue('test' in context.exception.args[0][0].args)

    def test_state_exception(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.state_model_with_e(Exception('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 1)
        self.assertTrue('test' in context.exception.args[0][0].args)

    def test_state_type_error(self):
        state = {torchbearer.X: None, torchbearer.MODEL: self.state_model_with_e(TypeError('test')), torchbearer.CRITERION: lambda state: MagicMock(),
                 torchbearer.OPTIMIZER: Mock(), torchbearer.CALLBACK_LIST: Mock(), torchbearer.BACKWARD_ARGS: {}}

        closure = self.closure

        with self.assertRaises(Exception) as context:
            closure(state)

        self.assertTrue(len(context.exception.args[0]) == 2)


"""
Base metrics are the base classes which represent the metrics supplied with torchbearer. They all use the
:func:`.default_for_key` decorator so that they can be accessed in the call to :class:`.torchbearer.Model` via the
following strings:

- '`acc`' or '`accuracy`': The :class:`.DefaultAccuracy` metric
- '`binary_acc`' or '`binary_accuracy`': The :class:`.BinaryAccuracy` metric
- '`cat_acc`' or '`cat_accuracy`': The :class:`.CategoricalAccuracy` metric
- '`top_5_acc`' or '`top_5_accuracy`': The :class:`.TopKCategoricalAccuracy` metric
- '`top_10_acc`' or '`top_10_accuracy`': The :class:`.TopKCategoricalAccuracy` metric with k=10
- '`mse`': The :class:`.MeanSquaredError` metric
- '`loss`': The :class:`.Loss` metric
- '`epoch`': The :class:`.Epoch` metric
- '`lr`': The :class:`.LR` metric
- '`roc_auc`' or '`roc_auc_score`': The :class:`.RocAucScore` metric
"""
import torch.nn as nn
import torch.nn.functional as F

import torchbearer
from torchbearer.metrics import default_for_key, Metric, CategoricalAccuracy, MeanSquaredError, BinaryAccuracy

try:
    __loss_map__ = {
        # NN
        nn.CrossEntropyLoss.__name__: CategoricalAccuracy,
        nn.NLLLoss.__name__: CategoricalAccuracy,
        nn.MSELoss.__name__: MeanSquaredError,
        nn.BCELoss.__name__: BinaryAccuracy,
        nn.BCEWithLogitsLoss.__name__:  BinaryAccuracy,
        # Functional
        F.cross_entropy.__name__: CategoricalAccuracy,
        F.nll_loss.__name__: CategoricalAccuracy,
        F.mse_loss.__name__: MeanSquaredError,
        F.binary_cross_entropy.__name__: BinaryAccuracy,
        F.binary_cross_entropy_with_logits.__name__: BinaryAccuracy
    }
except AttributeError:  # Thrown when building the docs with mocked pytorch
    __loss_map__ = {}


@default_for_key('accuracy')
@default_for_key('acc')
class DefaultAccuracy(Metric):
    """The default accuracy metric loads in a different accuracy metric depending on the loss function or criterion in
    use at the start of training. Default for keys: `acc`, `accuracy`. The following bindings are in place for both nn
    and functional variants:

    - cross entropy loss -> :class:`.CategoricalAccuracy` [DEFAULT]
    - nll loss -> :class:`.CategoricalAccuracy`
    - mse loss -> :class:`.MeanSquaredError`
    - bce loss -> :class:`.BinaryAccuracy`
    - bce loss with logits -> :class:`.BinaryAccuracy`
    """
    def __init__(self):
        super(DefaultAccuracy, self).__init__('placeholder')  # Don't set yet, wait for reset

        self.metric = CategoricalAccuracy()  # Default to CategoricalAccuracy
        self.name = self.metric.name
        self._loaded = False
        self._train = True

    def train(self):
        self._train = True
        return self.metric.train()

    def eval(self, data_key=None):
        self._train = False
        return self.metric.eval(data_key=data_key)

    def process(self, *args):
        return self.metric.process(*args)

    def process_final(self, *args):
        return self.metric.process_final(*args)

    def reset(self, state):
        if not self._loaded:
            criterion = state[torchbearer.CRITERION]

            name = None

            if hasattr(criterion, '__name__'):
                name = criterion.__name__
            elif hasattr(criterion, '__class__'):
                name = criterion.__class__.__name__

            if name is not None and name in __loss_map__:
                self.metric = __loss_map__[name]()
                self.name = self.metric.name
                if self._train:
                    self.metric.train()
                else:
                    self.metric.eval(data_key=state[torchbearer.DATA])

            self._loaded = True

        return self.metric.reset(state)



import unittest

import torch
from torch.autograd import Variable

import torchbearer
from torchbearer.metrics import Loss, Epoch, CategoricalAccuracy, TopKCategoricalAccuracy, BinaryAccuracy, MeanSquaredError


class TestLoss(unittest.TestCase):
    def setUp(self):
        with torch.no_grad():
            self._state = {
                torchbearer.LOSS: torch.FloatTensor([2.35])
            }
        self._metric = Loss().root  # Get root node of Tree for testing

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process(self._state)
        self.assertAlmostEqual(2.35, result[0], 3, 0.002)

    def test_validate_process(self):
        self._metric.eval()
        result = self._metric.process(self._state)
        self.assertAlmostEqual(2.35, result[0], 3, 0.002)


class TestEpoch(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.EPOCH: 101
        }
        self._metric = Epoch().metric  # Get wrapped metric for testing

    def test_process(self):
        result = self._metric.process(self._state)
        self.assertEqual(101, result)

    def test_process_final(self):
        result = self._metric.process_final(self._state)
        self.assertEqual(101, result)


class TestBinaryAccuracy(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: torch.LongTensor([
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [0, 0, 1],
                [0, 1, 0]
            ]),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1],  # Correct
                [0.1, 0.9, 0.1],  # Correct
                [0.1, 0.1, 0.9],  # Correct
                [0.9, 0.1, 0.1],  # Incorrect
                [0.9, 0.1, 0.1]  # Incorrect
            ])
        }
        self._targets = [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 1]
        self._metric = BinaryAccuracy().root  # Get root node of Tree for testing

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_validate_process(self):
        self._metric.eval()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_weird_types(self):
        state = {
            torchbearer.Y_TRUE: torch.LongTensor([
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [0, 0, 1],
                [0, 1, 0]
            ]).byte(),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1],  # Correct
                [0.1, 0.9, 0.1],  # Correct
                [0.1, 0.1, 0.9],  # Correct
                [0.9, 0.1, 0.1],  # Incorrect
                [0.9, 0.1, 0.1]  # Incorrect
            ]).double()
        }

        self._metric.train()
        result = self._metric.process(state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_threshold(self):
        state = {
            torchbearer.Y_TRUE: torch.FloatTensor([
                [0.9, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [0, 0.3, 1],
                [0, 0.6, 0]
            ]),
            torchbearer.Y_PRED: torch.FloatTensor([
                [0.9, 0.1, 0.1],  # Correct
                [0.1, 0.9, 0.1],  # Correct
                [0.1, 0.1, 0.9],  # Correct
                [0.9, 0.1, 0.1],  # Incorrect
                [0.9, 0.1, 0.1]  # Incorrect
            ])
        }

        metric = BinaryAccuracy(threshold=0.4).root  # Get root node of Tree for testing
        metric.train()
        result = metric.process(state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))


class TestCategoricalAccuracy(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: Variable(torch.LongTensor([0, 1, 2, 2, 1])),
            torchbearer.Y_PRED: Variable(torch.FloatTensor([
                [0.9, 0.1, 0.1], # Correct
                [0.1, 0.9, 0.1], # Correct
                [0.1, 0.1, 0.9], # Correct
                [0.9, 0.1, 0.1], # Incorrect
                [0.9, 0.1, 0.1], # Incorrect
            ])),
        }
        self._targets = [1, 1, 1, 0, 0]
        self._metric = CategoricalAccuracy().root  # Get root node of Tree for testing

    def test_ignore_index(self):
        metric = CategoricalAccuracy(ignore_index=1).root  # Get root node of Tree for testing
        targets = [1, 1, 0]

        metric.train()
        result = metric.process(self._state)
        for i in range(0, len(targets)):
            self.assertEqual(result[i], targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(targets[i])
                                 + ' in: ' + str(result))

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_train_process_soft(self):
        self._metric.train()
        soft_targets = torch.FloatTensor([
                [0.98, 0.01, 0.01], # Correct
                [0.01, 0.98, 0.01], # Correct
                [0.01, 0.01, 0.98], # Correct
                [0.01, 0.01, 0.98], # Incorrect
                [0.01, 0.98, 0.01], # Incorrect
            ])
        state = self._state.copy()
        state[torchbearer.Y_TRUE] = soft_targets
        result = self._metric.process(state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_validate_process(self):
        self._metric.eval()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))


class TestTopKCategoricalAccuracy(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: Variable(torch.LongTensor([0, 5, 2, 3, 1])),
            torchbearer.Y_PRED: Variable(torch.FloatTensor([
                [0.9, 0.8, 0.7, 0.6, 0.5, 0.4],  # Correct
                [0.4, 0.5, 0.6, 0.7, 0.8, 0.9],  # Correct
                [0.6, 0.5, 0.4, 0.7, 0.8, 0.9],  # Incorrect
                [0.6, 0.5, 0.7, 0.4, 0.8, 0.9],  # Incorrect
                [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]  # Correct
            ]))
        }
        self._targets = [1, 1, 0, 0, 1]
        self._metric = TopKCategoricalAccuracy(k=5).root  # Get root node of Tree for testing

    def test_ignore_index(self):
        metric = TopKCategoricalAccuracy(ignore_index=1).root  # Get root node of Tree for testing
        targets = [1, 1, 0, 0]

        metric.train()
        result = metric.process(self._state)
        for i in range(0, len(targets)):
            self.assertEqual(result[i], targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(targets[i])
                                 + ' in: ' + str(result))

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_validate_process(self):
        self._metric.eval()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertEqual(result[i], self._targets[i],
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_top_ten_default(self):
        metric = torchbearer.metrics.get_default('top_10_acc').root
        self.assertEqual(metric.k, 10)


class TestMeanSquaredError(unittest.TestCase):
    def setUp(self):
        self._state = {
            torchbearer.Y_TRUE: Variable(torch.FloatTensor(
                [0.8, 0.2, 0.0, 0.4, 0.3, 0.7]
            )),
            torchbearer.Y_PRED: Variable(torch.FloatTensor(
                [0.9, 0.1, 0.1, 0.7, 0.5, 0.6]
            ))
        }
        self._targets = [0.01, 0.01, 0.01, 0.09, 0.04, 0.01]
        self._metric = MeanSquaredError().root  # Get root node of Tree for testing

    def test_train_process(self):
        self._metric.train()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertAlmostEqual(result[i].item(), self._targets[i], places=3,
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))

    def test_validate_process(self):
        self._metric.eval()
        result = self._metric.process(self._state)
        for i in range(0, len(self._targets)):
            self.assertAlmostEqual(result[i].item(), self._targets[i], places=3,
                             msg='returned: ' + str(result[i]) + ' expected: ' + str(self._targets[i])
                                 + ' in: ' + str(result))


import torch
from unittest import TestCase

from mock import MagicMock, patch

import torchbearer
from torchbearer.callbacks import UnpackState


class TestUnpackState(TestCase):
    def reset_state(self):
        return {
            torchbearer.X: 'data',
            torchbearer.Y_TRUE: 'targets'
        }

    def test_sample(self):
        packer = UnpackState(keys=[torchbearer.X, torchbearer.Y_TRUE])
        state = self.reset_state()

        packer.on_sample(state)
        self.assertTrue(state[torchbearer.X] == {torchbearer.X: 'data', torchbearer.Y_TRUE: 'targets'})

        state = self.reset_state()
        packer.on_sample_validation(state)
        self.assertTrue(state[torchbearer.X] == {torchbearer.X: 'data', torchbearer.Y_TRUE: 'targets'})

    def test_no_key(self):
        packer = UnpackState()
        state = self.reset_state()

        packer.on_sample(state)
        self.assertTrue(state[torchbearer.X] == 'data')

        state = self.reset_state()
        packer.on_sample_validation(state)
        self.assertTrue(state[torchbearer.X] == 'data')

    def test_fill_X(self):
        packer = UnpackState(keys=[torchbearer.Y_TRUE])
        state = self.reset_state()

        packer.on_sample(state)
        self.assertTrue(state[torchbearer.X] == {torchbearer.X: 'data', torchbearer.Y_TRUE: 'targets'})

        state = self.reset_state()
        packer.on_sample_validation(state)
        self.assertTrue(state[torchbearer.X] == {torchbearer.X: 'data', torchbearer.Y_TRUE: 'targets'})

    def test_forward_no_dict(self):
        packer = UnpackState(keys=[torchbearer.Y_TRUE])
        state = self.reset_state()

        state[torchbearer.Y_PRED] = 1
        packer.on_forward(state)
        self.assertTrue(state[torchbearer.Y_PRED] == 1)

        state = self.reset_state()

        state[torchbearer.Y_PRED] = 1
        packer.on_forward_validation(state)
        self.assertTrue(state[torchbearer.Y_PRED] == 1)

    def test_forward_list(self):
        packer = UnpackState(keys=[torchbearer.Y_TRUE])
        state = self.reset_state()

        state[torchbearer.Y_PRED] = [1, 2, 3]
        packer.on_forward(state)
        self.assertTrue(state[torchbearer.Y_PRED] == [1, 2, 3])

        state = self.reset_state()

        state[torchbearer.Y_PRED] = [1, 2, 3]
        packer.on_forward_validation(state)
        self.assertTrue(state[torchbearer.Y_PRED] == [1, 2, 3])

    def test_forward_dict_no_y_pred(self):
        packer = UnpackState(keys=[torchbearer.Y_TRUE])
        state = self.reset_state()

        state[torchbearer.Y_PRED] = {'one': 1, 'two': 2}
        packer.on_forward(state)
        self.assertTrue(state[torchbearer.Y_PRED] == {'one': 1, 'two': 2})

        state = self.reset_state()

        state[torchbearer.Y_PRED] = {'one': 1, 'two': 2}
        packer.on_forward_validation(state)
        self.assertTrue(state[torchbearer.Y_PRED] == {'one': 1, 'two': 2})

    def test_forward_dict_y_pred(self):
        packer = UnpackState(keys=[torchbearer.Y_TRUE])
        state = self.reset_state()

        state[torchbearer.Y_PRED] = {torchbearer.Y_PRED: 1, 'two': 2}
        packer.on_forward(state)
        self.assertTrue(state[torchbearer.Y_PRED] == 1)
        self.assertTrue(state['two'] == 2)

        state = self.reset_state()

        state[torchbearer.Y_PRED] = {torchbearer.Y_PRED: 1, 'two': 2}
        packer.on_forward_validation(state)
        self.assertTrue(state[torchbearer.Y_PRED] == 1)
        self.assertTrue(state['two'] == 2)


from unittest import TestCase

import torch

import torchbearer
from torchbearer.callbacks import BCPlus


class TestBCPlus(TestCase):
    def test_on_val(self):
        bcplus = BCPlus(classes=4)
        state = {torchbearer.TARGET: torch.tensor([1, 3, 2])}
        bcplus.on_sample_validation(state)
        self.assertTrue((state[torchbearer.TARGET] -
                         torch.tensor([[0, 1, 0, 0],
                                       [0, 0, 0, 1],
                                       [0, 0, 1, 0]]).float()).abs().lt(1e-4).all())

        bcplus = BCPlus(classes=4)
        state = {torchbearer.TARGET: torch.tensor([[0, 1, 0, 0],
                                                   [0, 0, 0, 1],
                                                   [0, 0, 1, 0]])}
        bcplus.on_sample_validation(state)
        self.assertTrue((state[torchbearer.TARGET] -
                         torch.tensor([[0, 1, 0, 0],
                                       [0, 0, 0, 1],
                                       [0, 0, 1, 0]]).float()).abs().lt(1e-4).all())

    def test_bc_loss(self):
        prediction = torch.tensor([[10.0, 0.01]])
        target = torch.tensor([[0., 0.8]])
        loss = BCPlus.bc_loss({torchbearer.PREDICTION: prediction, torchbearer.TARGET: target})
        self.assertTrue((loss - 7.81).abs().le(1e-2).all())

    def test_sample_targets(self):
        # Test mixup
        bcplus = BCPlus(classes=4, mixup_loss=True)
        state = {torchbearer.INPUT: torch.zeros(3, 10, 10), torchbearer.TARGET: torch.tensor([1, 3, 2]), torchbearer.DEVICE: 'cpu'}
        bcplus.on_sample(state)

        self.assertTrue(torchbearer.MIXUP_LAMBDA in state)
        self.assertTrue(torchbearer.MIXUP_PERMUTATION in state)
        self.assertTrue(len(state[torchbearer.TARGET]) == 2)

        # Test bcplus
        bcplus = BCPlus(classes=4)
        state = {torchbearer.INPUT: torch.zeros(3, 10, 10), torchbearer.TARGET: torch.tensor([1, 3, 2]),
                 torchbearer.DEVICE: 'cpu'}
        bcplus.on_sample(state)

        self.assertTrue(state[torchbearer.TARGET].dim() == 2)
        self.assertTrue(not (torchbearer.MIXUP_PERMUTATION in state))

    def test_sample_inputs(self):
        torch.manual_seed(7)

        batch = torch.tensor([[
            [0.1, 0.5, 0.6],
            [0.8, 0.6, 0.5],
            [0.2, 0.4, 0.7]
        ]])
        target = torch.tensor([1])
        state = {torchbearer.INPUT: batch, torchbearer.TARGET: target, torchbearer.DEVICE: 'cpu'}

        bcplus = BCPlus(classes=4)
        bcplus.on_sample(state)

        lam = torch.ones(1) * 0.2649

        self.assertTrue(((state[torchbearer.INPUT] * (lam.pow(2) + (1 - lam).pow(2)).sqrt()) - (batch - batch.mean())).abs().le(1e-4).all())



from torchbearer import Metric
import warnings

__keys__ = []


def state_key(key):
    """Computes and returns a non-conflicting key for the state dictionary when given a seed key

    Args:
        key (str): The seed key - basis for new state key

    Returns:
        StateKey: New state key
    """
    return StateKey(key)


class StateKey(Metric):
    """ StateKey class that is a unique state key based on the input string key. State keys are also metrics which
    retrieve themselves from state.

    Args:
        key (str): Base key
    """
    def __init__(self, key):
        self.key = self._gen_key_(key)
        super(StateKey, self).__init__(self.key)

    def process(self, state):
        return {self.name: state[self]}

    def process_final(self, state):
        return {self.name: state[self]}

    def __call__(self, state):
        return state[self]

    def _gen_key_(self, key):
        if key in __keys__:
            count = 1
            my_key = key + '_' + str(count)

            while my_key in __keys__:
                count += 1
                my_key = key + '_' + str(count)

            key = my_key

        __keys__.append(key)
        return key

    def __repr__(self):
        return self.key

    def __str__(self):
        return self.key

    def __eq__(self, other):
        return self.key == str(other)

    def __hash__(self):
        return self.key.__hash__()


class State(dict):
    """
    State dictionary that behaves like a python dict but accepts StateKeys
    """
    def __init__(self):
        super(State, self).__init__()

    def get_key(self, statekey):
        if isinstance(statekey, str):
            warnings.warn("State was accessed with a string: {}, generate keys with StateKey(str).".format(statekey), stacklevel=3)
        return statekey

    @property
    def data(self):
        new_state = State()
        for key in self.keys():
            try:
                new_state[key] = self[key].data
            except AttributeError:
                new_state[key] = self[key]
        return new_state

    def __getitem__(self, key):
        return super(State, self).__getitem__(self.get_key(key))

    def __setitem__(self, key, val):
        super(State, self).__setitem__(self.get_key(key), val)

    def __delitem__(self, val):
        super(State, self).__delitem__(val)

    def __contains__(self, o):
        return super(State, self).__contains__(self.get_key(o))

    def update(self, d):
        new_dict = {}
        for key in d:
            new_dict[self.get_key(key)] = d[key]
        super(State, self).update(new_dict)


#: The torchbearer version
VERSION = state_key('torchbearer_version')

#: The PyTorch module / model that will be trained
MODEL = state_key('model')

#: The criterion to use when model fitting
CRITERION = state_key('criterion')

#: The optimizer to use when model fitting
OPTIMIZER = state_key('optimizer')

#: The device currently in use by the :class:`.Trial` and PyTorch model
DEVICE = state_key('device')

#: The data type of tensors in use by the model, match this to avoid type issues
DATA_TYPE = state_key('dtype')

#: The list of metrics in use by the :class:`.Trial`
METRIC_LIST = state_key('metric_list')

#: The metric dict from the current batch of data
METRICS = state_key('metrics')

#: A self refrence to the Trial object for persistence etc.
SELF = state_key('self')

#: The current epoch number
EPOCH = state_key('epoch')

#: The total number of epochs to run for
MAX_EPOCHS = state_key('max_epochs')

#: The string name of the current data
DATA = state_key('data')

#: The current data generator (DataLoader)
GENERATOR = state_key('generator')

#: The current iterator
ITERATOR = state_key('iterator')

#: The current number of steps per epoch
STEPS = state_key('steps')

#: The train data generator in the Trial object
TRAIN_GENERATOR = state_key('train_generator')

#: The number of train steps to take
TRAIN_STEPS = state_key('train_steps')

#: The flag representing train data
TRAIN_DATA = state_key('train_data')

#: Flag for refreshing of training iterator when finished instead of each epoch
INF_TRAIN_LOADING = state_key('inf_train_loading')

#: The validation data generator in the Trial object
VALIDATION_GENERATOR = state_key('validation_generator')

#: The number of validation steps to take
VALIDATION_STEPS = state_key('validation_steps')

#: The flag representing validation data
VALIDATION_DATA = state_key('validation_data')

#: The test data generator in the Trial object
TEST_GENERATOR = state_key('test_generator')

#: The number of test steps to take
TEST_STEPS = state_key('test_steps')

#: The flag representing test data
TEST_DATA = state_key('test_data')

#: A flag that can be set to true to stop the current fit call
STOP_TRAINING = state_key('stop_training')

#: The current batch of ground truth data
TARGET = Y_TRUE = state_key('y_true')

#: The current batch of predictions
PREDICTION = Y_PRED = state_key('y_pred')

#: The current batch of inputs
INPUT = X = state_key('x')

#: The sampler which loads data from the generator onto the correct device
SAMPLER = state_key('sampler')

#: The batch loader which handles formatting data from each batch
LOADER = state_key('loader')

#: The current value for the loss
LOSS = state_key('loss')

#: The key which maps to the predictions over the dataset when calling predict
FINAL_PREDICTIONS = state_key('final_predictions')

#: The current batch number
BATCH = state_key('t')

#: The timings keys used by the timer callback
TIMINGS = state_key('timings')

#: The :class:`.CallbackList` object which is called by the Trial
CALLBACK_LIST = state_key('callback_list')

#: The history list of the Trial instance
HISTORY = state_key('history')

#: The optional arguments which should be passed to the backward call
BACKWARD_ARGS = state_key('backward_args')

#: The lambda coefficient of the linear combination of inputs
MIXUP_LAMBDA = state_key('mixup_lambda')

#: The permutation of input indices for input mixup
MIXUP_PERMUTATION = state_key('mixup_permutation')



import torchbearer
from torchbearer import Callback
import torch
import torch.nn.functional as F
from torch.distributions import Beta

from torchbearer.bases import cite

bc = """
@inproceedings{tokozume2018between,
  title={Between-class learning for image classification},
  author={Tokozume, Yuji and Ushiku, Yoshitaka and Harada, Tatsuya},
  booktitle={Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition},
  pages={5486--5494},
  year={2018}
}
"""


@cite(bc)
class BCPlus(Callback):
    """BC+ callback which mixes images by treating them as waveforms. For standard BC, see :class:`.Mixup`.
    This callback can optionally convert labels to one hot before combining them according to the lambda parameters,
    sampled from a beta distribution, use alpha=1 to replicate the paper. Use with :meth:`BCPlus.bc_loss` or set
    `mixup_loss = True` and use :meth:`.Mixup.mixup_loss`.

    .. note::

       This callback first sets all images to have zero mean. Consider adding an offset (e.g. 0.5) back before
       visualising.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import BCPlus

        # Example Trial which does BCPlus regularisation
        >>> bcplus = BCPlus(classes=10)
        >>> trial = Trial(None, criterion=BCPlus.bc_loss, callbacks=[bcplus], metrics=['acc'])

    Args:
        mixup_loss (bool): If True, the lambda and targets will be stored for use with the mixup loss function.
        alpha (float): The alpha value for the beta distribution.
        classes (int): The number of classes for conversion to one hot.

    State Requirements:
        - :attr:`torchbearer.state.X`: State should have the current data stored and correctly normalised
        - :attr:`torchbearer.state.Y_TRUE`: State should have the current data stored
    """

    def __init__(self, mixup_loss=False, alpha=1, classes=-1):
        super(BCPlus, self).__init__()
        self.mixup_loss = mixup_loss
        self.classes = classes
        self.dist = Beta(torch.tensor([float(alpha)]), torch.tensor([float(alpha)]))

    @staticmethod
    def bc_loss(state):
        """The KL divergence between the outputs of the model and the ratio labels. Model ouputs should be un-normalised
        logits as this function performs a log_softmax.

        Args:
            state: The current :class:`Trial` state.
        """
        prediction, target = state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE]

        entropy = - (target[target.nonzero().split(1, dim=1)] * target[target.nonzero().split(1, dim=1)].log()).sum()
        cross = - (target * F.log_softmax(prediction, dim=1)).sum()

        return (cross - entropy) / prediction.size(0)

    def _to_one_hot(self, target):
        if target.dim() == 1:
            target = target.unsqueeze(1)
            one_hot = torch.zeros_like(target).repeat(1, self.classes)
            one_hot.scatter_(1, target, 1)
            return one_hot
        return target.float()

    def on_sample(self, state):
        super(BCPlus, self).on_sample(state)

        lam = self.dist.sample().to(state[torchbearer.DEVICE])

        permutation = torch.randperm(state[torchbearer.X].size(0))

        batch1 = state[torchbearer.X]
        batch1 = batch1 - batch1.view(batch1.size(0), -1).mean(1, keepdim=True).view(*tuple([batch1.size(0)] + [1] * (batch1.dim() - 1)))
        g1 = batch1.view(batch1.size(0), -1).std(1, keepdim=True).view(*tuple([batch1.size(0)] + [1] * (batch1.dim() - 1)))

        batch2 = batch1[permutation]
        g2 = g1[permutation]

        p = 1. / (1 + ((g1 / g2) * ((1 - lam) / lam)))

        state[torchbearer.X] = (batch1 * p + batch2 * (1 - p)) / (p.pow(2) + (1 - p).pow(2)).sqrt()

        if not self.mixup_loss:
            target = self._to_one_hot(state[torchbearer.TARGET]).float()
            state[torchbearer.Y_TRUE] = lam * target + (1 - lam) * target[permutation]
        else:
            state[torchbearer.MIXUP_LAMBDA] = lam
            state[torchbearer.MIXUP_PERMUTATION] = permutation
            state[torchbearer.Y_TRUE] = (state[torchbearer.Y_TRUE], state[torchbearer.Y_TRUE][state[torchbearer.MIXUP_PERMUTATION]])

    def on_sample_validation(self, state):
        super(BCPlus, self).on_sample_validation(state)
        if not self.mixup_loss:
            state[torchbearer.TARGET] = self._to_one_hot(state[torchbearer.TARGET]).float()
